import numpy as np
from .abstract_ranker import AbstractRanker

# An Explore-then-Commit strategy: similar to random until each playlist has been displayed n times or more,
# then recommends the top n_reco playlists with highest mean observed rewards, for each segment
class NaiveGreedy(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(NaiveGreedy, self).__init__(config, dataObj)
        # self.user_segment = user_segment
        # n_segments = len(np.unique(self.user_segment))
        self.ranking_display = np.zeros((dataObj.n_users, dataObj.n_items))
        self.ranking_success = np.zeros((dataObj.n_users, dataObj.n_items))
        self.min_n = int(parameters["min-n"]["value"])
        # self.cascade_model = cascade_model

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        # user_segment = np.take(self.user_segment, batch_users)
        user_success = np.take(self.ranking_success, batch_users, axis = 0)
        user_displays = np.take(self.ranking_display, batch_users, axis = 0).astype(float)
        user_random_score = np.random.random(user_displays.shape)
        user_score = np.divide(user_success, user_displays, out=np.zeros_like(user_displays), where=user_displays!=0)
        discounted_displays = np.maximum(np.zeros_like(user_displays), self.min_n - user_displays)
        user_choice = np.lexsort((user_random_score, - user_score, -discounted_displays))[:, :self.config.list_size]
        # Shuffle l_init first slots
        # np.random.shuffle(user_choice[0:l_init])
        return user_choice

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        batch_size = len(batch_users)
        for i in range(batch_size):
            # user_segment = self.user_segment[batch_users[i]]
            # total_stream = len(rewards[i].nonzero())
            nb_display = 0
            for r, c in zip(rankings[i], clicks[i]):
                nb_display += 1
                self.ranking_success[batch_users[i]][r] += c
                self.ranking_display[batch_users[i]][r] += 1
                # if self.cascade_model and ((total_stream == 0 and nb_display == l_init) or (r == 1)):
                #     break
        return


import numpy as np
import os
from pathlib import Path
import json
from fair_dynamic_rec.core.util.utils import get_legend_labels
from fair_dynamic_rec.core.util.utils import get_param_config_name
from scipy import stats

class MetricCmd:
    def __init__(self, config, dataObj, rankers):
        self.optimal_reward_filename = 'optimal-reward'
        self.ranker_reward_filename = 'ranker-reward'
        self.rankers = rankers
        config.get_metrics()
        self.load_rewards(config)

    def compute(self, config, dataObj, viz):
        for metric in config.metrics:
            if metric['name'] == 'n-step-regret':
                viz.draw(metric, self.n_step_regret(), {'x': 'rounds', 'y': metric, 'legend': get_legend_labels(self.rankers)}, 'line')
            if metric['name'] == 'cumulative-alpha-ia':
                viz.draw(metric, self.cumulative_alpha_ia(config, dataObj), {'x': 'rounds', 'y': metric, 'legend': get_legend_labels(self.rankers)}, 'line')
            if metric['name'] == 'item-exposure-per-round':
                viz.draw(metric, self.item_exposure_per_round(config, dataObj, metric), {'x': 'rounds', 'y': 'item rank', 'legend': get_legend_labels(self.rankers)}, 'scatter_colorbar')
            if metric['name'] == 'item-exposure-per-round-with-FPTP':
                viz.draw(metric, self.item_exposure_per_round_with_FPTP(config, dataObj, metric), {'x': 'rounds', 'y': 'item rank', 'legend': get_legend_labels(self.rankers)}, 'scatter_colorbar')
            if metric['name'] == 'ranking-popularity-correlation':
                viz.draw(metric, self.ranking_popularity_correlation(config, dataObj, metric), {'x': 'rounds', 'y': 'correlation', 'legend': get_legend_labels(self.rankers)}, 'scatter')
            if metric['name'] == 'pri':
                viz.draw(metric, self.PRI(config, dataObj, metric), {'x': 'rounds', 'y': 'correlation', 'legend': get_legend_labels(self.rankers)}, 'scatter')

    def n_step_regret(self):
        regret = {}
        for key, value in self.rewards.items():
            regret[key] = {'x':range(len(value['optimal'])), 'y': np.cumsum(np.array(value['optimal']) - np.array(value['ranker']))}
        return regret

    def cumulative_alpha_ia(self, config, dataObj):
        ia = self.compute_ia_per_round(config, dataObj)
        cumulative_ia = {}
        for key, value in ia.items():
            tmp = np.zeros(value.shape[0])
            for i in range(value.shape[0]):
                tmp[i] = np.count_nonzero(np.sum(value[:i+1, :], axis=0) >= 1) / dataObj.n_items
            cumulative_ia[key] = {'x': range(len(tmp)), 'y': tmp}
        return cumulative_ia

    def item_exposure_per_round(self, config, dataObj, metric):
        item_exposure = {}
        subdirs, dir_name = self.get_result_subdir(config)
        ranker_index = 0
        for subdir in subdirs:
            tmp = np.zeros((dataObj.n_items, config.rounds))
            for i in range(config.rounds):
                with open(str(subdir / str(i)), 'r') as f:
                    items_position = self.load_rankings_as_item_avg_ranking(config, dataObj, json.load(f))
                    tmp[:, i] = items_position
            item_exposure[dir_name[ranker_index]] = {'data': tmp}
                    # for item in range(items_position.shape[0]):
                    #     item_position = np.argwhere(items_position[item] > 0) + 1
                    #     item_exposure[dir_name[ranker_index]][item, i] = np.mean(item_position)
            if metric['orderby'] == 'ranking':
                item_exposure[dir_name[ranker_index]] = item_exposure[dir_name[ranker_index]][np.sum(item_exposure[dir_name[ranker_index]], axis=1).argsort()][::-1]
            elif metric['orderby'] == 'popularity':
                item_exposure[dir_name[ranker_index]] = item_exposure[dir_name[ranker_index]][self.compute_item_popularity(dataObj).argsort()][::-1]
            # item_exposure[dir_name[ranker_index]] = item_exposure[dir_name[ranker_index]][item_exposure[dir_name[ranker_index]][:,1].argsort()][::-1]
            ranker_index += 1
        return item_exposure

    def item_exposure_per_round_with_FPTP(self, config, dataObj, metric):
        item_exposure = {}
        subdirs, dir_name = self.get_result_subdir(config)
        ranker_index = 0
        for subdir in subdirs:
            tmp = np.zeros((dataObj.n_items, config.rounds))
            tmp_fp, tmp_tp = np.zeros((dataObj.n_items, config.rounds)), np.zeros((dataObj.n_items, config.rounds))
            for i in range(config.rounds):
                with open(str(subdir / str(i)), 'r') as f:
                    # items_position = self.load_ranking_as_item_position(config, dataObj, json.load(f))
                    items_position, positive_feedback, negative_feedback = self.load_ranking_pos_neg_feedback_as_matrix(config, dataObj, json.load(f))
                    tmp[:, i] = items_position
                    tmp_fp[:, i] = negative_feedback
                    tmp_tp[:, i] = positive_feedback
            item_exposure[dir_name[ranker_index]] = {'data': tmp, 'positive_feedback': tmp_tp, 'negative_feedback': tmp_fp}
            if metric['orderby'] == 'ranking':
                item_exposure[dir_name[ranker_index]]['data'] = item_exposure[dir_name[ranker_index]]['data'][np.sum(item_exposure[dir_name[ranker_index]], axis=1).argsort()][::-1]
                item_exposure[dir_name[ranker_index]]['positive_feedback'] = item_exposure[dir_name[ranker_index]]['positive_feedback'][np.sum(item_exposure[dir_name[ranker_index]], axis=1).argsort()][::-1]
                item_exposure[dir_name[ranker_index]]['negative_feedback'] = item_exposure[dir_name[ranker_index]]['negative_feedback'][np.sum(item_exposure[dir_name[ranker_index]], axis=1).argsort()][::-1]
            elif metric['orderby'] == 'popularity':
                item_exposure[dir_name[ranker_index]]['data'] = item_exposure[dir_name[ranker_index]]['data'][self.compute_item_popularity(dataObj).argsort()][::-1]
                item_exposure[dir_name[ranker_index]]['positive_feedback'] = item_exposure[dir_name[ranker_index]]['positive_feedback'][self.compute_item_popularity(dataObj).argsort()][::-1]
                item_exposure[dir_name[ranker_index]]['negative_feedback'] = item_exposure[dir_name[ranker_index]]['negative_feedback'][self.compute_item_popularity(dataObj).argsort()][::-1]
            ranker_index += 1
        return item_exposure

    def ranking_popularity_correlation(self, config, dataObj, metric):
        correlation = {}
        subdirs, dir_name = self.get_result_subdir(config)
        ranker_index = 0
        popularity = self.compute_item_popularity(dataObj)
        for subdir in subdirs:
            tmp = np.zeros(config.rounds)
            for i in range(config.rounds):
                with open(str(subdir / str(i)), 'r') as f:
                    items_position = self.load_rankings_as_item_avg_ranking(config, dataObj, json.load(f))
                    tmp[i] = -stats.spearmanr(items_position, popularity).correlation
            correlation[dir_name[ranker_index]] = {'x': range(len(tmp)), 'y': tmp}
            ranker_index += 1
        return correlation

    def PRI(self, config, dataObj, metric):
        correlation = {}
        subdirs, dir_name = self.get_result_subdir(config)
        ranker_index = 0
        popularity = self.compute_item_popularity(dataObj)
        for subdir in subdirs:
            tmp = np.zeros(config.rounds)
            for i in range(config.rounds):
                with open(str(subdir / str(i)), 'r') as f:
                    # test_data = dataObj.test_data[np.nonzero(np.any(dataObj.test_data != 0, axis=0))[0]]
                    items_position_matrix = self.load_rankings_as_item_ranking(config, dataObj, json.load(f))
                    avg_rank = np.multiply(dataObj.test_data, items_position_matrix).sum(axis=0) / np.count_nonzero(dataObj.test_data, axis=0)
                    tmp[i] = -stats.spearmanr(avg_rank[avg_rank.nonzero()[0]], popularity[avg_rank.nonzero()[0]]).correlation
            correlation[dir_name[ranker_index]] = {'x': range(len(tmp)), 'y': tmp}
            ranker_index += 1
        return correlation


    def compute_ia_per_round(self, config, dataObj):
        ia = {}
        subdirs, dir_name = self.get_result_subdir(config)
        ranker_index = 0
        for subdir in subdirs:
            ia[dir_name[ranker_index]] = np.zeros((config.rounds, dataObj.n_items))
            for i in range(config.rounds):
                with open(str(subdir / str(i)), 'r') as f:
                    ranking = self.load_ranking_as_matrix(dataObj, json.load(f))
                    ia[dir_name[ranker_index]][i, :] = np.sum(ranking, axis=0)
            ranker_index += 1
        return ia

    def load_rewards(self, config):
        self.rewards = {}
        subdirs, dir_name = self.get_result_subdir(config)
        i = 0
        for subdir in subdirs:
            with open(str(subdir / self.optimal_reward_filename), 'r') as f:
                optimal_reward = json.load(f)
            with open(str(subdir / self.ranker_reward_filename), 'r') as f:
                ranker_reward = json.load(f)
            self.rewards[dir_name[i]] = {'optimal': optimal_reward['reward'], 'ranker': ranker_reward['reward']}
            i += 1

    def get_result_subdir(self, config):
        subdirs, dir_name = [], []
        result_dir = config._target / Path('results')
        for i in range(len(self.rankers)):
            param_name = get_param_config_name(self.rankers[i]["config"])
            subdirs.append(result_dir / param_name)
            dir_name.append(param_name)

        # dirs = os.listdir(result_dir)
        # for dir in dirs:
        #     if os.path.isdir(result_dir / dir):
        #         subdirs.append(result_dir / dir)
        #         dir_name.append(dir)

        return subdirs, dir_name

    def compute_item_popularity(self, dataObj):
        return np.count_nonzero(dataObj.train_data, axis=0) / dataObj.train_data.shape[1]

    def load_ranking_as_matrix(self, dataObj, json_ranking):
        ranking = np.zeros((dataObj.n_users, dataObj.n_items))
        for key, value in json_ranking.items():
            ranking[dataObj.userid_mapped_data[key]][self.convert_itemids_to_internal_ids(dataObj, value['r'])] += 1
        return ranking
    def load_ranking_as_item_exposure_matrix(self, config, dataObj, json_ranking):
        items_position = np.zeros((dataObj.n_items, config.list_size))
        for key, value in json_ranking.items():
            ranked_items = self.convert_itemids_to_internal_ids(dataObj, value['r'])
            for i in range(len(ranked_items)):
                items_position[ranked_items[i]][i] += 1
        return items_position
    def load_rankings_as_item_ranking(self, config, dataObj, json_ranking):
        items_position = np.zeros((dataObj.n_users, dataObj.n_items))
        for key, value in json_ranking.items():
            ranked_items = self.convert_itemids_to_internal_ids(dataObj, value['r'])
            for i in range(len(ranked_items)):
                items_position[i][ranked_items[i]] += i + 1
        return items_position
    def load_rankings_as_item_avg_ranking(self, config, dataObj, json_ranking):
        items_position = np.zeros(dataObj.n_items)
        ranking_count = np.zeros(dataObj.n_items)
        for key, value in json_ranking.items():
            ranked_items = self.convert_itemids_to_internal_ids(dataObj, value['r'])
            for i in range(len(ranked_items)):
                items_position[ranked_items[i]] += i + 1
                ranking_count[ranked_items[i]] += 1
        np.seterr(invalid='ignore')
        avg_position = np.divide(items_position, ranking_count)
        avg_position[np.isnan(avg_position)] = 0
        return avg_position
    def load_ranking_pos_neg_feedback_as_matrix(self, config, dataObj, json_ranking):
        items_position = np.zeros(dataObj.n_items)
        ranking_count = np.zeros(dataObj.n_items)
        positive_feedback = np.zeros(dataObj.n_items)
        negative_feedback = np.zeros(dataObj.n_items)
        for key, value in json_ranking.items():
            ranked_items = self.convert_itemids_to_internal_ids(dataObj, value['r'])
            for i in range(len(ranked_items)):
                items_position[ranked_items[i]] += i + 1
                ranking_count[ranked_items[i]] += 1

            clicked_items = self.convert_itemids_to_internal_ids(dataObj, value['c'])
            previous_clicked_item_index = 0
            for clicked_item in clicked_items:
                positive_feedback[clicked_item] += 1
                current_clicked_item_index = ranked_items.index(clicked_item)
                negative_feedback[ranked_items[previous_clicked_item_index:current_clicked_item_index]] += 1
        np.seterr(invalid='ignore')
        avg_position = np.divide(items_position, ranking_count)
        avg_position[np.isnan(avg_position)] = 0
        return avg_position, positive_feedback, negative_feedback
    def load_pos_neg_feedback(self, config, dataObj, json_ranking):
        positive_feedback = np.zeros(dataObj.n_items)
        negative_feedback = np.zeros(dataObj.n_items)
        for key, value in json_ranking.items():
            ranked_items = self.convert_itemids_to_internal_ids(dataObj, value['r'])
            clicked_items = self.convert_itemids_to_internal_ids(dataObj, value['c'])
            previous_clicked_item_index = 0
            for clicked_item in clicked_items:
                positive_feedback[clicked_item] += 1
                current_clicked_item_index = ranked_items.index(clicked_item)
                negative_feedback[ranked_items[previous_clicked_item_index:current_clicked_item_index]] += 1
        return positive_feedback, negative_feedback

    def convert_itemids_to_internal_ids(self, dataObj, items):
        internal_list = []
        for item in items:
            internal_list.append(dataObj.itemid_mapped_data[item])
        return internal_list


import numpy as np
from .linear_submodular_bandit import LSB
from fair_dynamic_rec.core.util.utils import compute_user_K_prime

class EARSLSB(LSB):
    def __init__(self, config, dataObj, parameters=None):
        super(EARSLSB, self).__init__(config, dataObj, parameters)
        self.shuffle_K = int(parameters['shuffle_K']['value'])
        self.epsilon = float(parameters['epsilon']['value'])
        self.gamma = float(parameters['gamma']['value'])

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        """
        :param x: features
        :param k: number of positions
        :return: ranking: the ranked item id.
        delta: the conditional topic coverage of each item. Eq. (3) of NIPS 11 paper.
        """
        # assert x.shape[0] >= k
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.dim))
        tie_breaker = self.prng.rand(len(self.dataObj.feature_data['train_item_topical_features']))
        for i in range(len(batch_users)):
            user = batch_users[i]
            coverage = np.zeros(self.dim)
            ranking = []
            scores = []
            for j in range(self.config.list_size):
                # Line 8 - 11 of Nips 11
                gain_in_topic_coverage = self.conditional_coverage(x=self.dataObj.feature_data['train_item_topical_features'], coverage=coverage)
                cb = self.alpha * np.sqrt(np.multiply(np.dot(gain_in_topic_coverage, self.MInv[user]), gain_in_topic_coverage).sum(axis=1))
                score = np.dot(gain_in_topic_coverage, self.theta[user])
                ucb = score + cb + 1e-6 * tie_breaker

                winner = np.argmax(ucb)
                while winner in ranking:
                    ucb[winner] = -np.inf
                    winner = np.argmax(ucb)

                ranking.append(winner)
                self.batch_features[i][j] = gain_in_topic_coverage[winner]
                scores.append(ucb[winner])

                coverage = self.ranking_coverage(self.dataObj.feature_data['train_item_topical_features'][ranking])
            rankings[i] = np.asarray(self.shuffling_topK(ranking, scores, self.config.list_size))
        return rankings

    def shuffling_topK(self, ranking, scores, K):
        # Reshuffle top-12 if specified
        if self.shuffle_K > 1:
            np.random.shuffle(ranking[:self.shuffle_K])

        # Personalised reshuffling
        if self.shuffle_K == -1:
            # For every user
            # gamma = 0.9
            # E_c = Parallel(n_jobs=-1)(delayed(AbstractRanker.powerset_expectation_negation_partial)(scores, gamma, 1))
            # print(f'Expected clicks without shuffling top-K:', np.mean(E_c), np.var(E_c))
            # K = 6
            # E_c = Parallel(n_jobs=-1)(delayed(AbstractRanker.powerset_expectation_negation_partial)(scores, gamma, K))
            # print(f'Expected clicks after shuffling Top-{K}:', np.mean(E_c), np.var(E_c))

            # K_s = Parallel(n_jobs=-1)(delayed(compute_user_K_prime)(scores, self.gamma, K, epsilon=self.epsilon))
            K_s = compute_user_K_prime(scores, self.gamma, K, epsilon=self.epsilon)

            np.random.shuffle(ranking[:K_s])
        return ranking


import numpy as np
from .hybrid_lsb_linucb import HybridLSBLinUCB
from fair_dynamic_rec.core.util.utils import compute_user_K_prime

class EARSHybridLSBLinUCB(HybridLSBLinUCB):
    def __init__(self, config, dataObj, parameters=None):
        super(EARSHybridLSBLinUCB, self).__init__(config, dataObj, parameters)
        self.shuffle_K = int(parameters['shuffle_K']['value'])
        self.epsilon = float(parameters['epsilon']['value'])
        self.gamma = float(parameters['gamma']['value'])

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        """
        :param x: n * (n_topic + n_feature) np array
        :param k: number of positions
        :return: ranking: the ranked item id.
        """
        # assert x.shape[0] >= k
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.topic_dim + self.latent_dim))
        tie_breaker = self.prng.rand(len(self.dataObj.feature_data['train_item_latent_features']))

        # self.delta_t = np.zeros((k, x.shape[1]))
        # delta = x

        # z = delta[:, :self.n_z]  # topic
        # x = delta[:, self.n_z:]  # feature

        # tie_breaker = self.prng.rand(len(delta))
        for i in range(len(batch_users)):
            scores = []
            user = batch_users[i]
            # for CB
            BA_X = np.matmul(self.B[user].T, self.A_x_inv[user])  # B.T  A_x^-1
            ABA = np.matmul(self.A_z_inv[user], BA_X)  # A_z^-1 B.T A_X^-1
            ABABA = np.matmul(BA_X.T, ABA)

            # score and cb for the relevance
            score_x = np.dot(self.dataObj.feature_data['train_item_latent_features'], self.theta[user])
            XAX = np.multiply(np.dot(self.dataObj.feature_data['train_item_latent_features'], self.A_x_inv[user]), self.dataObj.feature_data['train_item_latent_features']).sum(axis=1)  # x^T A_X^-1 X^T
            cb_x = np.multiply(np.dot(self.dataObj.feature_data['train_item_latent_features'], ABABA), self.dataObj.feature_data['train_item_latent_features']).sum(axis=1) + XAX
            ucb_x = score_x + 1e-6 * tie_breaker
            ABAX = np.dot(ABA, self.dataObj.feature_data['train_item_latent_features'].T).T
            # score and cb for the topic
            # delta_t = []
            batch_features = []
            coverage = np.zeros(self.topic_dim)
            ranking = []
            # ranking_set = set()
            for j in range(self.config.list_size):
                # Line 8 - 11 of Nips 11
                z_t = self.conditional_coverage(x=self.dataObj.feature_data['train_item_topical_features'], coverage=coverage)
                ZAZ = np.multiply(np.dot(z_t, self.A_z_inv[user]), z_t).sum(axis=1)  # Z^T A_Z^-1 Z^T
                # cb_z = ZAZ - 2 * np.multiply(np.dot(z_t, ABA), x).sum(axis=1)
                cb_z = ZAZ - 2 * np.multiply(z_t, ABAX).sum(axis=1)

                # if (self.mitigation is None):
                cb = self.alpha * np.sqrt(cb_z + cb_x)
                # else:
                #     cb = self.alpha * (1 - (self.n_recommended / (t + 1))) * np.sqrt(cb_z + cb_x)
                score_z = np.dot(z_t, self.beta[i])
                ucb = ucb_x + score_z + cb
                # ucb = ucb_x + (1-user_avg_popularity)*score_z + cb

                # if(self.mitigation is not None):
                #     # tmp_n_recommended = self.n_recommended.copy()
                #     # tmp_n_recommended[np.where(tmp_n_recommended == 0)] = 1
                #     # ucb = (1-pow(self.n_recommended/(t+1),2)) * ucb
                #     ucb = (1 - (self.n_recommended / (t + 1))) * ucb

                winner = np.argmax(ucb)
                while winner in ranking:
                    ucb[winner] = -np.inf
                    winner = np.argmax(ucb)

                ranking.append(winner)
                # ranking_set.add(winner)
                batch_features.append(z_t[winner])

                scores.append(ucb[winner])

                coverage = self.ranking_coverage(self.dataObj.feature_data['train_item_topical_features'][ranking])

            rankings[i] = np.asarray(self.shuffling_topK(ranking, scores, self.config.list_size))
            self.batch_features[i][:, :self.topic_dim] = np.asarray(batch_features)
            self.batch_features[i][:, self.topic_dim:] = self.dataObj.feature_data['train_item_latent_features'][rankings[i]]
        return rankings

    def shuffling_topK(self, ranking, scores, K):
        # Reshuffle top-12 if specified
        if self.shuffle_K > 1:
            np.random.shuffle(ranking[:self.shuffle_K])

        # Personalised reshuffling
        if self.shuffle_K == -1:
            # For every user
            # gamma = 0.9
            # E_c = Parallel(n_jobs=-1)(delayed(AbstractRanker.powerset_expectation_negation_partial)(scores, gamma, 1))
            # print(f'Expected clicks without shuffling top-K:', np.mean(E_c), np.var(E_c))
            # K = 6
            # E_c = Parallel(n_jobs=-1)(delayed(AbstractRanker.powerset_expectation_negation_partial)(scores, gamma, K))
            # print(f'Expected clicks after shuffling Top-{K}:', np.mean(E_c), np.var(E_c))

            # K_s = Parallel(n_jobs=-1)(delayed(compute_user_K_prime)(scores, self.gamma, K, epsilon=self.epsilon))
            K_s = compute_user_K_prime(scores, self.gamma, K, epsilon=self.epsilon)

            np.random.shuffle(ranking[:K_s])
        return ranking


import numpy as np
from .abstract_ranker import AbstractRanker


# Segment-based Thompson Sampling strategy, with Beta(alpha_zero,beta_zero) priors
class ThompsonSampling(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(ThompsonSampling, self).__init__(config, dataObj)
        # self.user_segment = user_segment
        # n_segments = len(np.unique(self.user_segment))
        self.ranking_display = np.zeros((dataObj.n_users, dataObj.n_items))
        self.ranking_success = np.zeros((dataObj.n_users, dataObj.n_items))
        self.alpha = float(parameters["alpha"]["value"])
        self.beta = float(parameters["beta"]["value"])
        # self.t = 0
        # self.cascade_model = cascade_model

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        # user_segment = np.take(self.user_segment, batch_users)
        user_displays = np.take(self.ranking_display, batch_users, axis = 0).astype(float)
        user_success = np.take(self.ranking_success, batch_users, axis = 0)
        user_score = np.random.beta(self.alpha + user_success, self.beta + user_displays - user_success)
        user_choice = np.argsort(-user_score)[:, :self.config.list_size]
        # Shuffle l_init first slots
        # np.random.shuffle(user_choice[0:l_init])
        return user_choice

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        batch_size = len(batch_users)
        for i in range(batch_size):
            # user_segment = self.user_segment[user_ids[i]]
            # total_stream = len(rewards[i].nonzero())
            nb_display = 0
            for r, c in zip(rankings[i], clicks[i]):
                nb_display += 1
                self.ranking_success[batch_users[i]][r] += c
                self.ranking_display[batch_users[i]][r] += 1
                # if self.cascade_model and ((total_stream == 0 and nb_display == l_init) or (r == 1)):
                #     break
        return


import os
import numpy as np
from pathlib import Path
import matplotlib.pyplot as plt

class VisualizationCmd:
    def __init__(self, config, dataObj, rankers):
        self.viz_dir = 'visualizations'
        self.viz_path = self.make_viz_dir(config)

    def draw(self, metric, data, labels, plot_type):
        if plot_type == 'line':
            self.line_plot(metric, data, labels)
        elif plot_type == 'scatter':
            self.scatter_plot(metric, data, labels)
        elif plot_type == 'scatter_colorbar':
            self.scatter_colorbar_plot(metric, data, labels)
        return

    def line_plot(self, metric, data, labels):
        for key, value in data.items():
            plt.plot(value['x'], value['y'], label=labels['legend'][key])
        plt.xlabel(labels['x'])
        plt.ylabel(labels['y'])
        if len(data.keys()) > 1:
            plt.legend()
        plt.savefig(self.viz_path / Path(metric['name'] + '.eps'), format='eps')
        plt.close()

    def scatter_plot(self, metric, data, labels):
        for key, value in data.items():
            plt.scatter(value['x'], value['y'], label=labels['legend'][key], s=1)
        plt.xlabel(labels['x'])
        plt.ylabel(labels['y'])
        if len(data.keys()) > 1:
            plt.legend()
            # plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        plt.savefig(self.viz_path / Path(metric['name'] + '.eps'), format='eps')
        plt.close()

    def scatter_colorbar_plot(self, metric, data, labels):
        for key, value in data.items():
            heatmap = plt.pcolor(value['data'], cmap=plt.cm.Greys, vmin=1, vmax=10)

            # if 'positive_feedback' in value:
            #     for i in range(value['positive_feedback'].shape[1]):
            #         positive_items = value['positive_feedback'][:, i].nonzero()[0]
            #         plt.scatter(np.full(len(positive_items), i), positive_items, color=metric['TP-color'], s=0.1)
                    # negative_items = value['negative_feedback'][:, i].nonzero()[0]
                    # plt.scatter(np.full(len(negative_items), i), negative_items, color=metric['FP-color'], s=0.1)

            plt.yticks(np.arange(0, value['data'].shape[0], 100))
            plt.xticks(np.arange(0, value['data'].shape[1], value['data'].shape[1]/10))

            plt.colorbar(heatmap)

            plt.xlabel(labels['x'])
            plt.ylabel(labels['y'])

            plt.savefig(self.viz_path / Path(key + '-' + metric['name'] + '.jpg'), format='jpg')
            plt.close()

    def make_viz_dir(self, config):
        if not os.path.exists(config._target / Path(self.viz_dir)):
            os.makedirs(config._target / Path(self.viz_dir))
        return config._target / Path(self.viz_dir)


from scipy.optimize import minimize
import numpy as np


# Disclaimer: this class is taken from:
# https://gdmarmerola.github.io/ts-for-contextual-bandits/


# Defining a class for Online Bayesian Logistic Regression
class OnlineLogisticRegression:

    # Initializing
    def __init__(self, lambda_, alpha, n_dim, bias, maxiter=15):
        # Hyperparameter: deviation on the prior (L2 regularizer)
        self.lambda_ = lambda_
        self.alpha = alpha
        self.maxiter = maxiter

        # Initializing parameters of the model
        self.n_dim = n_dim
        # m: mean of the Bi, q inverse variance of the distribution
        self.m = np.zeros(self.n_dim)
        self.m[-1] = bias
        self.q = np.ones(self.n_dim) * self.lambda_

        # Initializing weights
        self.w = np.random.normal(self.m, self.alpha * (self.q) ** (-1.0), size=self.n_dim)

    # Loss function
    def loss(self, w, *args):
        X, y = args
        return 0.5 * (self.q[:-1] * (w[:-1] - self.m[:-1])).dot(w[:-1] - self.m[:-1]) + np.sum(
            [np.log(1 + np.exp(-y[j] * w.dot(X[j]))) for j in range(y.shape[0])])

    # Gradient
    def grad(self, w, *args):
        X, y = args
        return np.concatenate((self.q[:-1] * (w[:-1] - self.m[:-1]), 0.0), axis=None) + (-1) * np.array(
            [y[j] * X[j] / (1. + np.exp(y[j] * w.dot(X[j]))) for j in range(y.shape[0])]).sum(axis=0)

    # Fitting method
    def fit(self, X, y):
        # Step 1, find w
        self.w = minimize(self.loss, self.w, args=(X, y), jac=self.grad, method="L-BFGS-B",
                          options={'maxiter': self.maxiter}).x
        self.m = self.w

        # Step 2, update q
        P = (1 + np.exp(1 - X.dot(self.m))) ** (-1)
        self.q = self.q + (P * (1 - P)).dot(X ** 2)



import numpy as np

"""
The base class of all rankers. 
"""
class AbstractRanker():
    def __init__(self, config, dataObj):
        self.config= config
        self.dataObj = dataObj
        self.rankers = config.get_ranker_params()

    @staticmethod
    def get_features_for_optimal_ranker(config, dataObj):
        item_topic = dataObj.feature_data['train_item_topical_features'] if 'item_topic' in config.known_variables else dataObj.feature_data['test_item_topical_features']
        item_latent = dataObj.feature_data['train_item_latent_features'] if 'item_latent' in config.known_variables else dataObj.feature_data['train_item_latent_features']

        if config.full_feature:
            return np.concatenate((dataObj.feature_data['test_user_latent_features'], dataObj.feature_data['test_user_topical_features']), axis=1), np.concatenate((item_latent, item_topic), axis=1)
        elif config.optimal_ranker == 'naive_relevance' or config.optimal_ranker == 'sigmoid_relevance':
            return dataObj.feature_data['test_user_latent_features'], item_latent#dataObj.feature_data['train_item_latent_features']
        elif config.optimal_ranker == 'topical_coverage':
            return dataObj.feature_data['test_user_topical_features'], item_topic#dataObj.feature_data['test_item_topical_features']
        elif config.optimal_ranker == 'hybrid':
            # item_full_feature = np.zeros((dataObj.feature_data['train_item_latent_features'].shape[0], dataObj.feature_data['train_item_latent_features'].shape[1] + dataObj.feature_data['train_item_topical_features'].shape[1]))
            # item_full_feature[:, :dataObj.feature_data['train_item_topical_features'].shape[1]] = dataObj.feature_data['train_item_topical_features']
            # item_full_feature[:, dataObj.feature_data['train_item_topical_features'].shape[1]:] = dataObj.feature_data['train_item_latent_features']
            item_full_feature = np.zeros((item_latent.shape[0], item_latent.shape[1] + item_topic.shape[1]))
            item_full_feature[:, :item_topic.shape[1]] = item_topic
            item_full_feature[:, item_topic.shape[1]:] = item_latent
            user_full_feature = np.zeros((dataObj.feature_data['test_user_latent_features'].shape[0], dataObj.feature_data['test_user_latent_features'].shape[1] + dataObj.feature_data['test_user_topical_features'].shape[1]))
            for i in range(user_full_feature.shape[0]):
                user_full_feature[i,:] = np.concatenate(((1 - float(config.lmbda)) * dataObj.feature_data['test_user_topical_features'][i,:], float(config.lmbda) * dataObj.feature_data['test_user_latent_features'][i,:]))
            return user_full_feature, item_full_feature
        # elif config.optimal_ranker == 'ears_hybrid':
        #     item_full_feature = np.zeros((dataObj.feature_data['train_item_latent_features'].shape[0], dataObj.feature_data['train_item_latent_features'].shape[1] + dataObj.feature_data['train_item_topical_features'].shape[1]))
        #     item_full_feature[:, :dataObj.feature_data['train_item_topical_features'].shape[1]] = dataObj.feature_data['test_item_topical_features']
        #     item_full_feature[:, dataObj.feature_data['train_item_topical_features'].shape[1]:] = dataObj.feature_data['test_item_latent_features']
        #     user_full_feature = np.zeros((dataObj.feature_data['test_user_latent_features'].shape[0], dataObj.feature_data['test_user_latent_features'].shape[1] + dataObj.feature_data['test_user_topical_features'].shape[1]))
        #     for i in range(user_full_feature.shape[0]):
        #         user_full_feature[i,:] = np.concatenate(((1 - float(config.lmbda)) * dataObj.feature_data['train_user_topical_features'][i,:], float(config.lmbda) * dataObj.feature_data['train_user_latent_features'][i,:]))
        #     return user_full_feature, item_full_feature

    @staticmethod
    def ranking_coverage(s):
        """
        Return the coverage of an list s for topics. Eq 2 of Nips-11
        :param s: ranked list s in (0, 1), n by d numpy ndarray
        :return:
        """
        s = np.asarray(s) if type(s) is list else s
        return 1 - np.prod(1 - s, axis=0)

    @staticmethod
    def conditional_coverage(x, coverage):
        """
        Return the coverage of an item given the current ranking
        Based on Eq. 2 and Eq. 3 of NIPS11
        :param x: coverage of this item
        :param coverage: topic covergate of previous items
        :return: conditional coverage of x given ranking
        """
        x = np.asarray(x)
        coverage = np.asarray(coverage)
        return 1 - np.multiply(1 - x, 1 - coverage) - coverage



import numpy as np
from .abstract_ranker import AbstractRanker
import math
from fair_dynamic_rec.core.util.utils import compute_user_K_prime

class LSB(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(LSB, self).__init__(config, dataObj)
        self.n_samples = np.zeros(dataObj.n_users)
        self.n_clicks = np.zeros(dataObj.n_users)
        self.dim = self.dataObj.feature_data['train_item_topical_features'].shape[1]
        self.prng = np.random.RandomState(seed=config.seed)
        self.alpha = float(parameters["alpha"]["value"])

        self.sigma = float(parameters["sigma"]["value"])
        # self.t = 1
        # self.seed = seed
        # parameters
        self.ill_matrix_counter = 0
        self.theta = np.ones((self.dataObj.n_users, self.dim))  # d-dimensional
        self.b = np.zeros((self.dataObj.n_users, self.dim))  # d
        self.M = np.zeros((self.dataObj.n_users, self.dim, self.dim))  # d by d
        self.MInv = np.zeros((self.dataObj.n_users, self.dim, self.dim))  # for fast matrix inverse computation, d by d
        for i in range(self.dataObj.n_users):
            self.M[i] = np.eye(self.dim)
            self.MInv[i] = np.eye(self.dim)

        # for ill inverse
        self.b_tmp = np.zeros((self.dataObj.n_users, self.dim))
        self.MInv_tmp = np.zeros((self.dataObj.n_users, self.dim, self.dim))
        self.batch_features = None

        # parameters for feature_weight
        self.gamma = float(parameters.get("gamma",{}).get("value",0))
        self.processing_type = parameters.get("processing_type",{}).get("value",'')
        self.window = int(parameters.get('window',{}).get('value',0))
        self.click_history = np.zeros((self.dataObj.n_users, self.dim))

        # parameters for EARS
        self.shuffle_K = int(parameters.get('shuffle_K',{}).get('value',0))
        self.epsilon = float(parameters.get('epsilon',{}).get('value',0))
        self.ears_gamma = float(parameters.get('ears_gamma',{}).get('value',0))

        # parameters for item_weight
        self.item_coef = np.ones((self.dataObj.n_users, self.dataObj.n_items))

        # parameters for discount factor
        self.exp_recommended = np.zeros((self.dataObj.n_users, self.dataObj.n_items))
        self.exp_examined = np.zeros((self.dataObj.n_users, self.dataObj.n_items))

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        """
        :param x: features
        :param k: number of positions
        :return: ranking: the ranked item id.
        delta: the conditional topic coverage of each item. Eq. (3) of NIPS 11 paper.
        """
        # assert x.shape[0] >= k
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.dim))
        tie_breaker = self.prng.rand(len(self.dataObj.feature_data['train_item_topical_features']))
        for i in range(len(batch_users)):
            user = batch_users[i]
            coverage = np.zeros(self.dim)
            ranking = []
            scores = []
            for j in range(self.config.list_size):
                # Line 8 - 11 of Nips 11
                gain_in_topic_coverage = self.conditional_coverage(x=self.dataObj.feature_data['train_item_topical_features'], coverage=coverage)
                x = gain_in_topic_coverage
                if self.processing_type == 'item_weight':
                    x = self.item_coef[user].reshape(self.dataObj.n_items, 1) * gain_in_topic_coverage

                if self.processing_type == 'recommended_discountfactor':
                    cb = self.alpha * (1 - (self.exp_recommended[user] / (round + 1))) * np.sqrt(np.multiply(np.dot(x, self.MInv[user]), x).sum(axis=1))
                elif self.processing_type == 'examined_discountfactor':
                    cb = self.alpha * (1 - (self.exp_examined[user] / (round + 1))) * np.sqrt(np.multiply(np.dot(x, self.MInv[user]), x).sum(axis=1))
                else:
                    cb = self.alpha * np.sqrt(np.multiply(np.dot(x, self.MInv[user]), x).sum(axis=1))
                score = np.dot(x, self.theta[user])
                ucb = score + cb + 1e-6 * tie_breaker

                winner = np.argmax(ucb)
                while winner in ranking:
                    ucb[winner] = -np.inf
                    winner = np.argmax(ucb)

                ranking.append(winner)
                self.batch_features[i][j] = gain_in_topic_coverage[winner]
                scores.append(ucb[winner])

                coverage = self.ranking_coverage(self.dataObj.feature_data['train_item_topical_features'][ranking])

            if self.processing_type == 'EARS':
                rankings[i] = np.asarray(self.shuffling_topK(ranking, scores, self.config.list_size))
            else:
                rankings[i] = np.asarray(ranking)
        return rankings

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        for i in range(len(batch_users)):
            user = batch_users[i]
            _clicks, _batch_features = self.__collect_feedback(clicks, i)

            discount_coef = [1 / (math.log(1 + j)) for j in range(1, len(rankings[0]) + 1)]
            discount_coef_reward = [math.log(1+j) for j in range(1, len(_clicks) + 1)]
            discount_coef_penalization = [self.gamma * 1 / (math.log(1+j)) for j in range(1, len(_clicks) + 1)]

            if self.processing_type == 'recommended_discountfactor':
                self.exp_recommended[user][np.array(rankings[0])] += discount_coef
            elif self.processing_type == 'examined_discountfactor':
                if len(clicks) == 0:
                    self.exp_examined[user][np.array(rankings[0])] += discount_coef
                else:
                    self.exp_examined[user][np.array(rankings[0][:len(clicks)])] += discount_coef[:len(clicks)]

            if self.processing_type == 'item_weight':
                _batch_features = self.update_item_weight(rankings[0], _batch_features, _clicks, discount_coef_penalization, discount_coef_reward, user, user_round)

            """
            This is for computing self.theta (Line 3-5 of Alogirthm 1 of NIPS 11)
            For fast matrix inverse, we use Woodbury matrix identity (https://en.wikipedia.org/wiki/Woodbury_matrix_identity)

            Return: self.theta is updated.
            """
            # for the inverse of M, feature matrix
            # x * m^-1 * x^T
            xmx = np.dot(_batch_features, np.dot(self.MInv[user], _batch_features.T))
            # (1/sigma I + xmx)^-1
            try:
                tmp_inv = np.linalg.inv(1 / self.sigma * np.eye(len(_batch_features)) + xmx)
            except np.linalg.LinAlgError:
                # for the ill matrix. if the matrix is not invertible, we ignore this update
                self.ill_matrix_counter += 1
                return
            # m^-1*x^T
            MInv_xT = self.MInv[user].dot(_batch_features.T)
            # MInv_xT*tmp_inv*MInv_xT^T
            self.MInv_tmp[user] = np.dot(np.dot(MInv_xT, tmp_inv), MInv_xT.T)
            # MInv - the new part
            self.MInv[user] -= self.MInv_tmp[user]

            self.M[user] += self.sigma * _batch_features.T.dot(_batch_features)

            # for b: feedback
            if self.processing_type == 'feature_weight':
                self.update_feature_weight(_batch_features, _clicks, discount_coef_penalization, discount_coef_reward, user, user_round)
            else:
                self.b[user] += np.dot(_clicks, _batch_features)


            # self.b_tmp[user] = np.dot(_clicks, _batch_features)
            # self.b[user] += self.b_tmp[user]

            # for parameter theta
            self.theta[user] = np.dot(self.MInv[user], self.b[user])
            # self.theta[self.theta < 0] = 0

            self.n_samples[user] += len(_clicks)
            self.n_clicks[user] += sum(_clicks)

    def __collect_feedback(self, clicks, batch_user_id):
        """
        :param y:
        :return: the last observed position.
        """
        # With  Cascade assumption, only the first click counts.
        if self.config.feedback_model == 'cascade':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id], self.batch_features[batch_user_id]
            first_click = np.where(clicks[batch_user_id])[0][0]
            return clicks[batch_user_id][:first_click + 1], self.batch_features[batch_user_id][:first_click + 1]
        elif self.config.feedback_model == 'dcm':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id], self.batch_features[batch_user_id]
            last_click = np.where(clicks[batch_user_id])[0][-1]
            return clicks[batch_user_id][:last_click + 1], self.batch_features[batch_user_id][:last_click + 1]
        # all items are observed
        else:
            return clicks[batch_user_id], self.batch_features[batch_user_id]

    def shuffling_topK(self, ranking, scores, K):
        # Reshuffle top-12 if specified
        if self.shuffle_K > 1:
            np.random.shuffle(ranking[:self.shuffle_K])

        # Personalised reshuffling
        if self.shuffle_K == -1:
            # For every user
            # gamma = 0.9
            # E_c = Parallel(n_jobs=-1)(delayed(AbstractRanker.powerset_expectation_negation_partial)(scores, gamma, 1))
            # print(f'Expected clicks without shuffling top-K:', np.mean(E_c), np.var(E_c))
            # K = 6
            # E_c = Parallel(n_jobs=-1)(delayed(AbstractRanker.powerset_expectation_negation_partial)(scores, gamma, K))
            # print(f'Expected clicks after shuffling Top-{K}:', np.mean(E_c), np.var(E_c))

            # K_s = Parallel(n_jobs=-1)(delayed(compute_user_K_prime)(scores, self.gamma, K, epsilon=self.epsilon))
            K_s = compute_user_K_prime(scores, self.ears_gamma, K, epsilon=self.epsilon)

            np.random.shuffle(ranking[:K_s])
        return ranking

    def update_feature_weight(self, _batch_features, _clicks, discount_coef_penalization, discount_coef_reward, user, user_round):
        clicked_items_index = np.where(_clicks == 1)[0]
        _x = _batch_features
        if len(clicked_items_index) == 0:
            _x = -np.multiply(np.array(discount_coef_penalization).reshape(len(_clicks), 1), _x)
        else:
            previous_clicked_item_index = 0
            for clicked_item_index in clicked_items_index:
                current_clicked_item_index = clicked_item_index
                _x[current_clicked_item_index, :] = discount_coef_reward[current_clicked_item_index] * _x[current_clicked_item_index, :]
                _x[previous_clicked_item_index: current_clicked_item_index, :] = -np.multiply(np.array(discount_coef_penalization[previous_clicked_item_index: current_clicked_item_index]).reshape(current_clicked_item_index - previous_clicked_item_index, 1), _x[previous_clicked_item_index: current_clicked_item_index,:])
                previous_clicked_item_index = current_clicked_item_index + 1
        self.click_history[user] += _x.sum(axis=0)
        if user_round[user] % self.window == 0:
            self.b[user] += self.click_history[user] / self.window
            self.click_history[user] = np.zeros(self.dim)
        else:
            self.b[user] = np.dot(_clicks, _batch_features)
    def update_item_weight(self, rankings, _batch_features, _clicks, discount_coef_penalization, discount_coef_reward, user, user_round):
        clicked_items_index = np.where(_clicks == 1)[0]

        _x = _batch_features
        if len(clicked_items_index) == 0:
            self.item_coef[user][rankings] += -np.array(discount_coef_penalization)
            _x = self.item_coef[user][rankings][:len(_clicks)].reshape(len(_clicks),1) * _x
            # _x = -np.multiply(np.array(discount_coef_penalization).reshape(len(_clicks), 1), _batch_features)
        else:
            previous_clicked_item_index = 0
            for clicked_item_index in clicked_items_index:
                current_clicked_item_index = clicked_item_index
                self.item_coef[user][rankings[current_clicked_item_index]] += discount_coef_reward[current_clicked_item_index]
                _x[current_clicked_item_index,:] = self.item_coef[user][rankings[current_clicked_item_index]] * _x[current_clicked_item_index,:]
                # _x[current_clicked_item_index, :] = discount_coef_reward[current_clicked_item_index] * _batch_features[current_clicked_item_index, :]
                if current_clicked_item_index != previous_clicked_item_index:
                    self.item_coef[user][rankings[previous_clicked_item_index:current_clicked_item_index]] += -np.array(discount_coef_penalization[previous_clicked_item_index:current_clicked_item_index])
                    _x[previous_clicked_item_index:current_clicked_item_index,:] = self.item_coef[user][rankings[previous_clicked_item_index:current_clicked_item_index]].reshape(current_clicked_item_index - previous_clicked_item_index, 1) * _x[previous_clicked_item_index: current_clicked_item_index,:]
                # _x[previous_clicked_item_index: current_clicked_item_index, :] = -np.multiply(np.array(discount_coef_penalization[previous_clicked_item_index: current_clicked_item_index]).reshape(current_clicked_item_index - previous_clicked_item_index, 1), _batch_features[previous_clicked_item_index: current_clicked_item_index,:])
                previous_clicked_item_index = current_clicked_item_index + 1
        # self.click_history[user] += _x.sum(axis=0)
        if user_round[user] % self.window == 0:
            return _x
            # self.b[user] += self.click_history[user] / self.window
            # self.click_history[user] = np.zeros(self.dim)
        else:
            return _batch_features


import numpy as np
from .abstract_ranker import AbstractRanker
import pandas as pd
from fair_dynamic_rec.core.util.outputs import make_output_dir
from pathlib import Path
from sklearn.metrics.pairwise import cosine_similarity
import os.path

class NeighborUCB1(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(NeighborUCB1, self).__init__(config, dataObj)
        self.n_samples = np.zeros(dataObj.n_users)
        self.n_clicks = np.zeros(dataObj.n_users)
        self.prng = np.random.RandomState(seed=config.seed)
        self.sigma = float(parameters["sigma"]["value"]) if "sigma" in parameters else 1.0

        self.l = int(parameters["latent_dim"]["value"]) if "latent_dim" in parameters else 0
        self.lambda_1 = float(parameters["lambda"]["value"]) if "lambda" in parameters else 1.0
        self.alpha_v = float(parameters["alpha_v"]["value"]) if "alpha_v" in parameters else 1.0
        self.alpha_u = float(parameters["alpha_u"]["value"]) if "alpha_u" in parameters else 1.0
        self.noise = float(parameters["noise"]["value"]) if "noise" in parameters else 0.00001

        self.X = dataObj.train_data
        self.X = (self.X > 0).astype(float)
        self.X = self.X + self.noise

        self.XTX = np.dot(self.X.T, self.X)

        self.U = np.zeros((self.dataObj.n_items, self.l)) + self.noise
        self.V = np.zeros((self.dataObj.n_items, self.l)) + self.noise

        # A = U_T * (X_T.X + lambda) * U      -> l * l
        self.A = np.eye(self.l)
        self.AInv = np.linalg.inv(self.A)

        # U_T * (X_T . X_{u,i} - dMat(\etha))   -> l * m
        self.b = np.zeros((self.l, dataObj.n_items))

        # X_T . X + \lambda                   -> m * m
        self.C = self.XTX + self.lambda_1 * np.eye(dataObj.n_items)
        self.CInv = np.linalg.inv(self.C)

        # (X_T . X + dMat(\etha)) * V          -> m * l
        self.d = np.zeros((dataObj.n_items, self.l))

        self.E = np.eye(self.l)
        self.EInv = np.linalg.inv(self.E)

        self.writeSimMat()

        # self.ill_matrix_counter = 0
        # # for ill inverse
        # self.AInv_tmp = np.zeros((self.dataObj.n_users, (self.k + self.l) * self.dataObj.n_users, (self.k + self.l) * self.dataObj.n_users))
        # self.b_tmp = np.zeros((self.dataObj.n_users, (self.k + self.l) * self.dataObj.n_users))
        # self.CInv_tmp = np.zeros((self.dataObj.n_items, self.l, self.l))
        # self.d_tmp = np.zeros((self.dataObj.n_items, self.l))

    def get_ranking(self, batch_users, sampled_items=None, round=None):
        """
        :param x: features
        :param k: number of positions
        :return: ranking: the ranked item id.
        """
        # assert x.shape[0] >= k
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        # self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.dim))
        tie_breaker = self.prng.rand(self.dataObj.n_items)
        for i in range(len(batch_users)):
            user = batch_users[i]

            user_vector = np.reshape(self.X[user], (1,self.X.shape[1]))

            score = np.dot(self.X[user], np.dot(self.U, self.V.T))
            score[np.isnan(score)] = 0
            XU = np.multiply(user_vector.T,self.U) # 1 * k
            cb1 = np.sqrt(np.sum(np.multiply(np.dot(XU, self.AInv), XU), axis=1))
            cb1[np.isnan(cb1)] = 0
            cb2 = np.sqrt(np.multiply(np.dot(self.X[user], self.CInv), self.X[user]))
            cb2[np.isnan(cb2)] = 0
            cb3 = np.sqrt(np.sum(np.multiply(np.dot(self.V, self.EInv), self.V), axis=1))
            cb3[np.isnan(cb3)] = 0

            ucb = score + self.alpha_u * cb1 + self.alpha_v * cb2 * cb3

            rankings[i] = np.lexsort((tie_breaker, -ucb))[:self.config.list_size]

            self.writeCB(round+i, cb1, cb2, cb3)

        return rankings

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        for i in range(len(batch_users)):
            user = batch_users[i]

            _clicks, _ranking = self.__collect_feedback(clicks[i], rankings[i])

            clicked_items = _ranking[np.where(_clicks)[0]]
            X_clicked_items = np.zeros((self.dataObj.n_items, self.dataObj.n_users))
            if sum(_clicks) > 0:
                X_clicked_items[clicked_items, user] = np.ones((len(clicked_items),1))
                # Update XTX
                self.XTX = self.XTX + np.dot(X_clicked_items, self.X)

                # Update C and CInv
                self.C += np.dot(X_clicked_items, self.X)
                self.CInv = np.linalg.inv(self.C)

                # Update X
                self.X[user, clicked_items] += np.ones(len(clicked_items), dtype=float)

            # compute eta for U
            # eta = (np.dot(np.dot(np.dot(np.dot(np.dot(self.C,self.X[user,:].T),self.X[user,:]),self.V[_ranking]),self.E),self.V[_ranking].T)).diagonal() \
            #       / np.divide(self.C[_ranking,_ranking],np.dot(np.dot(self.V[_ranking],self.E),self.V[_ranking].T)).diagonal()
            eta_tmp = (np.dot(np.dot(np.dot(np.dot(np.dot(self.C, self.X.T), self.X), self.V), self.E),self.V.T)).diagonal() \
                  / np.divide(self.C,np.dot(np.dot(self.V, self.E), self.V.T))
            eta_tmp[np.isnan(eta_tmp)] = 0
            eta = np.zeros((self.dataObj.n_items, self.dataObj.n_items))
            eta = np.diag(np.diag(eta_tmp))[_ranking]

            # Update d
            self.d += np.dot((np.dot(self.X[user].T.reshape(self.dataObj.n_items,1), _clicks.reshape(1,len(_clicks))) - eta.T), self.V[_ranking])

            # Update U
            self.U = np.dot(np.dot(self.CInv, self.d), self.EInv)
            self.U = self.U / np.sqrt(np.sum(self.U ** 2))

            #compute eta for V
            # eta = np.dot(np.dot(np.dot(np.dot(self.U, self.AInv), self.U.T), self.X[user].T), X_clicked_items.T[user]).diagonal() \
            #       / np.dot(np.dot(self.U, self.AInv), self.U.T).diagonal()
            eta_tmp = np.dot(np.dot(np.dot(np.dot(self.U, self.AInv), self.U.T), self.X.T), self.X) \
                  / np.dot(np.dot(self.U, self.AInv), self.U.T)
            eta_tmp[np.isnan(eta_tmp)] = 0
            eta = np.diag(np.diag(eta_tmp))[_ranking]

            # Update A
            user_vector = np.reshape(self.X[user], (1, self.X.shape[1]))
            XU = np.multiply(user_vector.T, self.U)  # m * k
            self.A += np.dot(XU.T, XU) + self.lambda_1 * np.dot(self.U.T, self.U)
            try:
                self.AInv = np.linalg.inv(self.A)
            except np.linalg.LinAlgError:
                # for the ill matrix. if the matrix is not invertible, we ignore this update
                print('ill matrix A.')
                self.AInv = np.linalg.pinv(self.A)

            # Update b
            self.b[:, _ranking] += np.dot(self.U.T, (np.dot(self.X[user].T.reshape(self.dataObj.n_items,1), _clicks.reshape(1,len(_clicks))) - eta.T))

            # Update V.T
            self.V[_ranking] = np.dot(self.AInv, self.b[:, _ranking]).T
            # self.V = self.V / np.sqrt(np.sum(self.V ** 2))

            # Update E
            self.E = self.E + np.dot(self.V[_ranking].T, self.V[_ranking])
            self.EInv = np.linalg.inv(self.E)

            # Update U again
            self.U = np.dot(np.dot(self.CInv, self.d), self.EInv)
            self.U = self.U / np.sqrt(np.sum(self.U ** 2))

            self.n_samples[user] += len(_clicks)
            self.n_clicks[user] += sum(_clicks)

            self.writeParams(round)

    def __collect_feedback(self, click, ranking):
        """
        :param y:
        :return: the last observed position.
        """
        # With  Cascade assumption, only the first click counts.
        if self.config.feedback_model == 'cascade':
            if np.sum(click) == 0:
                return click, ranking
            first_click = np.where(click)[0][0]
            return click[:first_click + 1], ranking[:first_click +1]
        elif self.config.feedback_model == 'dcm':
            if np.sum(click) == 0:
                return click, ranking
            last_click = np.where(click)[0][-1]
            return click[:last_click + 1], ranking[:last_click + 1]
        # all items are observed
        else:
            return click, ranking

    def writeCB(self, round, cb1, cb2, cb3):
        f=open('cb1.txt','a')
        f.write(str(round)+'\t'+",".join(map(str, cb1)))
        f.close()

        f = open('cb2.txt', 'a')
        f.write(str(round) + '\t' + ",".join(map(str, cb2)))
        f.close()

        f = open('cb3.txt', 'a')
        f.write(str(round) + '\t' + ",".join(map(str, cb3)))
        f.close()

    def writeParams(self, round):
        if round == 0:
            return
        pre_path = self.config._target / Path('results')
        U_path = pre_path / Path('U'+str(round)+'.txt')
        V_path = pre_path / Path('V'+str(round)+'.txt')

        df = pd.DataFrame(data=self.U.astype(float))
        df.to_csv(U_path, sep=' ', header=False, float_format='%.10f', index=False)

        df = pd.DataFrame(data=self.V.astype(float))
        df.to_csv(V_path, sep=' ', header=False, float_format='%.10f', index=False)

    def writeSimMat(self):
        pre_path = self.config._target

        test_sim_path = pre_path / Path('test_item_sim.txt')
        train_sim_path = pre_path / Path('train_item_sim.txt')
        data_sim_path = pre_path / Path('data_item_sim.txt')

        if not os.path.isfile(test_sim_path):
            item_sim_matrix = cosine_similarity(self.dataObj.test_data.T)
            np.fill_diagonal(item_sim_matrix, 0)
            df = pd.DataFrame(data=item_sim_matrix.astype(float))
            df.to_csv(test_sim_path, sep=' ', header=False, float_format='%.10f', index=False)

        if not os.path.isfile(train_sim_path):
            item_sim_matrix = cosine_similarity(self.dataObj.train_data.T)
            np.fill_diagonal(item_sim_matrix, 0)
            df = pd.DataFrame(data=item_sim_matrix.astype(float))
            df.to_csv(train_sim_path, sep=' ', header=False, float_format='%.10f', index=False)

        if not os.path.isfile(data_sim_path):
            item_sim_matrix = cosine_similarity((self.dataObj.test_data + self.dataObj.train_data).T)
            np.fill_diagonal(item_sim_matrix, 0)
            df = pd.DataFrame(data=item_sim_matrix.astype(float))
            df.to_csv(data_sim_path, sep=' ', header=False, float_format='%.10f', index=False)


import numpy as np
from .linear_submodular_bandit import LSB
from .random import Random

class LSBRandom(LSB):
    def __init__(self, config, dataObj, parameters=None):
        super(LSBRandom, self).__init__(config, dataObj, parameters)
        self.initial_randomization_round = int(parameters['initial-randomization-round']['value'])
        self.random_ranker = Random(config, dataObj, None)

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        if round < self.initial_randomization_round:
            return self.random_ranker.get_ranking(batch_users, round)
        else:
            return super(LSBRandom, self).get_ranking(batch_users, round) #self.get_ranking(batch_users, round)

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        if round < self.initial_randomization_round:
            return self.random_ranker.update(batch_users, rankings, clicks, round)
        else:
            return super(LSBRandom, self).update(batch_users, rankings, clicks, round) #self.update(batch_users, rankings, clicks, round)


import numpy as np

class CascadeModel():
    name = 'CM'

    def __init__(self, config):
        # super(CascadeModel, self).__init__(*args, **kwargs)
        self.prng = np.random.RandomState(seed=config.seed)
        # assert np.linalg.norm(theta_star) <= 1.00001
        self.config = config
        self.coins = None
        self.coins_ready = False

    def set_coins(self, list_size):
        """
        -- call this, if you want to use the same coins for each positions each round
        -- set coins for each round.
        -- this should be called before get_feedback
        """
        self.coins = self.prng.rand(list_size)
        self.coins_ready = True

    def get_feedback(self, scores):
        # if type(delta) is list:
        #     delta = np.asarray(delta)
        # delta = np.tile(delta, (1, 1))
        # assert delta.shape[1] == self.theta.shape[0]

        # if not self.coins_ready:
        #     self.set_coins(self.config.list_size)
        #     self.del_coins()
        # score = np.dot(delta, self.theta)
        scores[scores > 1] = 1.
        # coins = score >= self.coins
        clicks, rewards = np.zeros((scores.shape)), np.zeros(scores.shape[0])
        n_users = scores.shape[0]
        for i in range(n_users):
            if not self.coins_ready:
                self.set_coins(self.config.list_size)
                self.del_coins()
            clicks[i] = scores[i] >= self.coins
            first_click = np.where(clicks[i])[0]
            if len(first_click) > 0:
                clicks[i][first_click[0]+1:] = False
            rewards[i] = sum(clicks[i])
        return clicks, rewards

    def del_coins(self):
        """
        This should be called by simulator at the end of each round.
        :return:
        """
        self.coins_ready = False

    def ave_feedback(self, delta):
        score = np.dot(delta, self.theta)
        score[score > 1] = 1.
        return 1 - np.prod(1-score)


import re
from datetime import datetime
import glob
import os
import numpy as np
from itertools import chain, combinations
from scipy import special

def create_log_name(filename: str):
    _time = str(datetime.now())
    _time_obj = datetime.strptime(_time, '%Y-%m-%d %H:%M:%S.%f')
    _timestamp = _time_obj.strftime("%Y%m%d_%H%M%S")
    return filename.format(_timestamp)

def purge_old_logs(path: str):
    for file in glob.glob(path):
        if re.match(r'.*/FairDynamicRec_log.*', file):
            os.remove(file)

def get_param_config_name(parameters):
    param_name = parameters['name']['attr']['abb'] if parameters['name']['attr']['abb'] else parameters['name']['value']
    for key, value in parameters.items():
        if key != 'name':
            param_name += "-" + str(parameters[key]['attr']['abb'] if parameters[key]['attr']['abb'] else parameters[key]['value']) + str(parameters[key]['value'])
    return param_name
def get_legend_labels(rankers):
    labels = {}
    for ranker in rankers:
        param_name = get_param_config_name(ranker['config'])
        labels[param_name] = param_name
        if 'attr' in ranker['config']['name']:
            if 'viz-legend' in ranker['config']['name']['attr']:
                if ranker['config']['name']['attr']['viz-legend'].lower() == 'false':
                    labels[param_name] = ranker['config']['name']['attr']['abb']
    return labels

def get_dict_key_from_value(var_dict, value):
    return list(var_dict.keys())[list(var_dict.values()).index(value)]
def get_dict_keys_from_list(var_dict, value_list):
    result = []
    for value in value_list:
        result.append(get_dict_key_from_value(var_dict, value))
    return result

def powerset_expectation_negation_partial(rankings, gamma, K):
    ''' Expected number of clicks over all possible permutations of R by counting powersets. '''

    N = len(rankings)
    rankings_neg = (1 - np.array(rankings)).tolist()

    result = 0
    for subset in powerset(rankings_neg[:K], K):
        k = len(subset)
        prod = np.multiply.reduce(subset)
        discount = gamma ** k
        if k != N:
            discount *= (1 - gamma)
        result += prod / special.binom(K, k) * discount

    for i in range(K, N):
        discount = gamma ** (i + 1)
        if i != (N - 1):
            discount *= (1 - gamma)
        result += np.multiply.reduce(rankings_neg[:i + 1]) * discount

    return 1 - result

def powerset(iterable, maxsize):
    ''' From https://stackoverflow.com/questions/1482308/how-to-get-all-subsets-of-a-set-powerset '''
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(maxsize + 1))

def compute_user_K_prime(rankings, gamma=0.9, max_K=12, epsilon=0.02):
    ''' Compute the value of K' for a user, up to which we can shuffle with max epsilon loss in clicks '''
    E_c = np.array([powerset_expectation_negation_partial(rankings, gamma, K) for K in range(1, max_K + 1)])
    E_c /= E_c[0]
    return np.searchsorted(-E_c, -(1.0 - epsilon))


import numpy as np
from .linear_submodular_bandit import LSB

class LinUCB(LSB):
    def __init__(self, config, dataObj, parameters=None):
        super(LinUCB, self).__init__(config, dataObj, parameters)
        self.dim = self.dataObj.feature_data['train_item_latent_features'].shape[1]

        # parameters
        self.ill_matrix_counter = 0
        self.theta = np.ones((self.dataObj.n_users, self.dim))  # d-dimensional
        self.b = np.zeros((self.dataObj.n_users, self.dim))  # d
        self.M = np.zeros((self.dataObj.n_users, self.dim, self.dim))  # d by d
        self.MInv = np.zeros((self.dataObj.n_users, self.dim, self.dim))  # for fast matrix inverse computation, d by d
        for i in range(self.dataObj.n_users):
            self.M[i] = np.eye(self.dim)
            self.MInv[i] = np.eye(self.dim)
        # for ill inverse
        self.b_tmp = np.zeros((self.dataObj.n_users, self.dim))
        self.MInv_tmp = np.zeros((self.dataObj.n_users, self.dim, self.dim))

        # self.gamma = float(parameters.get("gamma",{}).get("value",0))
        # self.window = int(parameters.get('window',{}).get('value',0))
        # self.click_history = np.zeros((self.dataObj.n_users, self.dim))
        #
        # self.shuffle_K = int(parameters.get('shuffle_K',{}).get('value',0))
        # self.epsilon = float(parameters.get('epsilon', {}).get('value', 0))
        # self.ears_gamma = float(parameters.get('ears_gamma', {}).get('value', 0))

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        """
        :param x: features
        :param k: number of positions
        :return: ranking: the ranked item id.
        """
        # assert x.shape[0] >= k
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.dim))
        tie_breaker = self.prng.rand(len(self.dataObj.feature_data['train_item_latent_features']))
        for i in range(len(batch_users)):
            user = batch_users[i]
            x = self.dataObj.feature_data['train_item_latent_features']
            if self.processing_type == 'item_weight':
                x = self.item_coef[user].reshape(self.dataObj.n_items, 1) * x

            if self.processing_type == 'recommended_discountfactor':
                cb = self.alpha * (1 - (self.exp_recommended[user] / (round + 1))) * np.sqrt(np.multiply(np.dot(x, self.MInv[user]), x).sum(axis=1))
            elif self.processing_type == 'examined_discountfactor':
                cb = self.alpha * (1 - (self.exp_examined[user] / (round + 1))) * np.sqrt(np.multiply(np.dot(x, self.MInv[user]), x).sum(axis=1))
            else:
                cb = self.alpha * np.sqrt(np.multiply(np.dot(x, self.MInv[user]), x).sum(axis=1))
            score = np.dot(x, self.theta[user])
            ucb = score + cb
            rankings[i] = np.lexsort((tie_breaker, -ucb))[:self.config.list_size]
            if self.processing_type == 'EARS':
                rankings[i] = np.asarray(self.shuffling_topK(rankings[i], np.sort(-ucb)[:self.config.list_size], self.config.list_size))
            self.batch_features[i] = self.dataObj.feature_data['train_item_latent_features'][rankings[i]]
        return rankings


from fair_dynamic_rec.core.util.errors import InvalidConfiguration
from fair_dynamic_rec.core.util.files import Files
from fair_dynamic_rec.core.util.xml_utils import xml_load_from_path
from pathlib import Path
import itertools

class ConfigCmd:
    """
    Loads the configuration file, inserts appropriate library contents,
    identifies the parameter variations and creates separate configurations
    for all combinations.
    """

    # _PARAM_NAME_PATH_RE = ".+\[@name='(.+)'\]"
    # _CV_DIR_RE = "cv_\d+"

    def __init__(self, config_file, target, log_filename=None):
        self._files = Files()
        self._target = target
        # self._log_filename = log_filename
        # self._count = None
        #
        self._files.set_study_path(target)
        self._files.set_config_file(config_file)
        # self._files.create_temp_dir()

        self._xml_input = self.read_xml(self._files.get_config_file_path())
        if self._xml_input is None:
            raise InvalidConfiguration('Parsing of config file failed.')
        # xml_data.findall('rec-alg/item-reg/value')
        # self.protected_features = ProtectedFeature(ProtectedFeature.parse_protected(self), self._files.get_temp_dir_path())
        #
        # if self.is_valid():
        #     self._files.set_data_path(self._xml_input)
        #     self._var_coll = VarColl()
        #     self._libraries = LibraryColl()
        #     self._bbo_steps = 0
        #
        #     self._key_password = None

        # self.ranker_config = self.get_rankers()

        self.seed = self.get_random_seed()

        self.processor_count = self.get_processor_count()

        self.known_variables = self.get_known_variables()

        self.rankers = self.get_ranker_params()
        self.simulator_name = self.get_simulator()
        if self.simulator_name == 'onlinesimulator':
            self.get_online_simulator_settings()

    def get_processor_count(self):
        if self._xml_input.findall('processor-count'):
            return int(self._xml_input.findall('processor-count')[0].text)
        else:
            return 1

    def get_target(self):
        return self._target

    def set_target(self, target):
        self._target = target

    def get_random_seed(self):
        return int(self._xml_input.findall('random-seed')[0].text)

    def get_general(self):
        return

    def get_data_path(self):
        return self.get_target() / Path(self._xml_input.findall('data/data-dir')[0].text)
    def get_rating_file_path(self):
        if self._xml_input.findall('data/rating-file'):
            return self.get_target() / Path(self._xml_input.findall('data/data-dir')[0].text) / self._xml_input.findall('data/rating-file')[0].text
        return None
    def get_train_file_path(self):
        if self._xml_input.findall('data/train-file'):
            return self.get_target() / Path(self._xml_input.findall('data/data-dir')[0].text) / self._xml_input.findall('data/train-file')[0].text
        return None
    def get_test_file_path(self):
        if self._xml_input.findall('data/test-file'):
            return self.get_target() / Path(self._xml_input.findall('data/data-dir')[0].text) / self._xml_input.findall('data/test-file')[0].text
        return None
    def get_user_file_path(self):
        if self._xml_input.findall('data/user-file'):
            return self.get_target() / Path(self._xml_input.findall('data/data-dir')[0].text) / self._xml_input.findall('data/user-file')[0].text
        return None
    def get_item_file_path(self):
        if self._xml_input.findall('data/item-file'):
            return self.get_target() / Path(self._xml_input.findall('data/data-dir')[0].text) / self._xml_input.findall('data/item-file')[0].text
        return None
    def get_supplier_file_path(self):
        if self._xml_input.findall('data/supplier-file'):
            return self.get_target() / Path(self._xml_input.findall('data/data-dir')[0].text) / self._xml_input.findall('data/supplier-file')[0].text
        return None

    def get_rating_data_delimiter(self):
        return str.encode(self._xml_input.findall('data/rating-file')[0].attrib['delimiter']).decode("unicode_escape")
    def get_train_data_delimiter(self):
        return str.encode(self._xml_input.findall('data/train-file')[0].attrib['delimiter']).decode("unicode_escape")
    def get_test_data_delimiter(self):
        return str.encode(self._xml_input.findall('data/test-file')[0].attrib['delimiter']).decode("unicode_escape")
    def get_user_data_delimiter(self):
        return str.encode(self._xml_input.findall('data/user-file')[0].attrib['delimiter']).decode("unicode_escape")
    def get_item_data_delimiter(self):
        return str.encode(self._xml_input.findall('data/item-file')[0].attrib['delimiter']).decode("unicode_escape")
    def get_supplier_data_delimiter(self):
        return str.encode(self._xml_input.findall('data/supplier-file')[0].attrib['delimiter']).decode("unicode_escape")

    def get_rating_data_size(self):
        n_rows, n_cols = None, None
        if self._xml_input.findall('data/rating-file'):
            if 'n-rows' in self._xml_input.findall('data/rating-file')[0].attrib:
                n_rows = int(self._xml_input.findall('data/rating-file')[0].attrib['n-rows'])
            if 'n-cols' in self._xml_input.findall('data/rating-file')[0].attrib:
                n_cols = int(self._xml_input.findall('data/rating-file')[0].attrib['n-cols'])
        return n_rows, n_cols

    def get_item_data_category_column(self):
        return int(self._xml_input.findall('data/item-file')[0].attrib['category_column'])
    def get_item_data_category_delimiter(self):
        return self._xml_input.findall('data/item-file')[0].attrib['category_delimiter']

    def get_splitter_model(self):
        splitter_model = 'splittraintest'
        if self._xml_input.findall('splitter/model'):
            splitter_model = self._xml_input.findall('splitter/model')[0].text
        return splitter_model
    def get_sampling_model(self):
        sampling_model = None
        if self._xml_input.findall('splitter/sampling'):
            sampling_model = self._xml_input.findall('splitter/sampling')[0].text
            if 'max-user' in self._xml_input.findall('splitter/sampling')[0].attrib:
                self.max_user = int(self._xml_input.findall('splitter/sampling')[0].attrib['max-user'])
            if 'max-item' in self._xml_input.findall('splitter/sampling')[0].attrib:
                self.max_item = int(self._xml_input.findall('splitter/sampling')[0].attrib['max-item'])
        return sampling_model
    def get_splitter_model_traintest_ratio(self):
        ratio = 0.5
        if self._xml_input.findall('splitter/model'):
            ratio = float(self._xml_input.findall('splitter/model')[0].attrib['ratio'])
        return ratio
    def get_data_binarization(self):
        binarize = False
        one_if_greater_than = 0
        if self._xml_input.findall('splitter/binarize'):
            binarize = self._xml_input.findall('splitter/binarize')[0].text == 'True'
            if 'one-if-greater-than' in self._xml_input.findall('splitter/binarize')[0].attrib:
                one_if_greater_than = float(self._xml_input.findall('splitter/binarize')[0].attrib['one-if-greater-than'])
        return binarize, one_if_greater_than

    def get_save_txt_data_info(self):
        save_data = False
        train_file_name, test_file_name = 'train.txt', 'test.txt'
        if self._xml_input.findall('splitter/save-txt'):
            save_data = self._xml_input.findall('splitter/save-txt')[0].text == 'True'
            train_file_name = None
            test_file_name = None
            if 'train-file-name' in self._xml_input.findall('splitter/save-txt')[0].attrib:
                train_file_name = self._xml_input.findall('splitter/save-txt')[0].attrib['train-file-name']
            if 'test-file-name' in self._xml_input.findall('splitter/save-txt')[0].attrib:
                test_file_name = self._xml_input.findall('splitter/save-txt')[0].attrib['test-file-name']
        return save_data, train_file_name, test_file_name
    def get_save_pkl_data_info(self):
        save_data = False
        file_name = 'data.pkl'
        if self._xml_input.findall('splitter/save-pkl'):
            save_data = self._xml_input.findall('splitter/save-pkl')[0].text
            file_name = self._xml_input.findall('splitter/save-pkl')[0].attrib['file-name']
        return save_data, file_name

    def get_ranker_params(self):
        rankers = []
        ranker_params = self._xml_input.findall('rankers/ranker')
        for ranker_param in ranker_params:
            ranker_single_val_params = {}
            _ranker_multiple_val_params = {}
            for child in ranker_param:
                if len(child) == 0:
                    ranker_single_val_params[child.tag] = {'value': child.text, 'attr': child.attrib}
                else:
                    _ranker_multiple_val_params[child.tag] = []
                    for i in range(len(child)):
                        _ranker_multiple_val_params[child.tag].append({'value': child.findall('value')[i].text, 'attr': child.attrib})

            ranker_multiple_val_params = {}
            if _ranker_multiple_val_params:
                keys, values = zip(*_ranker_multiple_val_params.items())
                ranker_multiple_val_params = [dict(zip(keys, v)) for v in itertools.product(*values)]
            rankers.append({'single_val_params': ranker_single_val_params, 'multiple_val_params': ranker_multiple_val_params})
        return rankers

    def get_simulator(self):
        return self._xml_input.findall('simulator/name')[0].text
    def get_online_simulator_settings(self):
        self.rounds = 1000
        if self._xml_input.findall('simulator/rounds'):
            self.rounds = int(self._xml_input.findall('simulator/rounds')[0].text)

        self.n_users_per_round = 100
        if self._xml_input.findall('simulator/n-users-per-round'):
            self.n_users_per_round = int(self._xml_input.findall('simulator/n-users-per-round')[0].text)

        self.delayed_update = False
        if self._xml_input.findall('simulator/delayed-update'):
            self.delayed_update = self._xml_input.findall('simulator/delayed-update')[0].text

        self.feedback_model = False
        if self._xml_input.findall('simulator/feedback-model'):
            self.feedback_model = self._xml_input.findall('simulator/feedback-model')[0].text

        self.optimal_ranker = 'naive_relevance'
        self.full_feature = False
        if self._xml_input.findall('simulator/optimal-ranker'):
            self.optimal_ranker = self._xml_input.findall('simulator/optimal-ranker')[0].text
            if 'full-feature' in self._xml_input.findall('simulator/optimal-ranker')[0].attrib:
                self.full_feature = self._xml_input.findall('simulator/optimal-ranker')[0].attrib['full-feature']
            if 'lmbda' in self._xml_input.findall('simulator/optimal-ranker')[0].attrib:
                self.lmbda = float(self._xml_input.findall('simulator/optimal-ranker')[0].attrib['lmbda'])

        self.optimal_reward_function = 'naive_relevance'
        if self._xml_input.findall('simulator/optimal-reward-function'):
            self.optimal_reward_function = self._xml_input.findall('simulator/optimal-reward-function')[0].text

        self.list_size = 10
        if self._xml_input.findall('simulator/list-size'):
            self.list_size = int(self._xml_input.findall('simulator/list-size')[0].text)
    def get_optimal_ranker(self):
        self.optimal_ranker = 'naive_relevance'
        self.full_feature = False
        if self._xml_input.findall('simulator/optimal-ranker'):
            self.optimal_ranker = self._xml_input.findall('simulator/optimal-ranker')[0].text
            if self._xml_input.findall('simulator/optimal-ranker')[0].attrib['full-feature'] is not None:
                self.full_feature = self._xml_input.findall('simulator/optimal-ranker')[0].attrib['full-feature']

    def get_latent_feature_dim(self):
        latent_feature_dim = 10
        if self._xml_input.findall('features/latent_feature_dim'):
            latent_feature_dim = int(self._xml_input.findall('features/latent_feature_dim')[0].text)
        return latent_feature_dim
    def normalize_latent_feature(self):
        normalize_latent_feature = False
        if self._xml_input.findall('features/latent_feature_normalize'):
            normalize_latent_feature = self._xml_input.findall('features/latent_feature_normalize')[0].text == 'True'
        return normalize_latent_feature
    def save_item_features(self):
        save_features = False
        if self._xml_input.findall('features/save-item-feature'):
            save_features = self._xml_input.findall('features/save-item-feature')[0].text == 'True'
        return save_features
    def get_features_derivation_type(self):
        features_derivation_type = []
        if self._xml_input.findall('features/derivation-type'):
            if len(self._xml_input.findall('features/derivation-type')) > 1:
                for child in self._xml_input.findall('features/derivation-type'):
                    features_derivation_type.append(child.findall('value').text.strip())
            else:
                features_derivation_type.append(self._xml_input.findall('features/derivation-type')[0].text.strip())
        return features_derivation_type
    def get_known_variables(self):
        known_variables = []
        if self._xml_input.findall('features/known-variables'):
            if len(self._xml_input.findall('features/known-variables')) > 1:
                for child in self._xml_input.findall('features/known-variables'):
                    known_variables.append(child.findall('value').text.strip())
            else:
                known_variables.append(self._xml_input.findall('features/known-variables')[0].text.strip())
        return known_variables

    def get_metrics(self):
        self.metrics = []
        for child in self._xml_input.findall('metric/value'):
            _dict = {'name': child.text}
            for key, value in child.attrib.items():
                _dict[key] = value
            self.metrics.append(_dict)

    # def get_xml(self):
    #     return self._xml_input
    #
    # def get_librec_auto_log_name(self):
    #     return self._log_filename
    #
    # # def get_var_data(self):
    # #     return self._var_data
    #
    # def get_key_password(self):
    #     return self._key_password
    #
    # def set_key_password(self, pw):
    #     self._key_password = pw
    #
    # def get_value_conf(self, subexp_no):
    #     return self._var_coll.var_confs[subexp_no]
    #
    # def has_alg_script(self):
    #     alg_script_elem = single_xpath(self._xml_input, '/librec-auto/alg/script')
    #     return (alg_script_elem is not None)
    #
    # def get_bbo_steps(self):
    #     if self._bbo_steps > 0:
    #         return self._bbo_steps
    #     else:
    #         return None
    #
    # def get_sub_exp_count(self):
    #     exp_count = len(self._var_coll.var_confs)
    #
    #     if self.get_bbo_steps() is not None:
    #         self._count = self.get_bbo_steps()
    #
    #     if self._count is not None:
    #         return self._count
    #     elif exp_count == 0:
    #         return 1
    #     else:
    #         return exp_count
    #
    # def get_files(self):
    #     return self._files

    def read_xml(self, path_str):
        path = self._files.get_config_file_path()
        if (path.exists()):
            xml_input = xml_load_from_path(path)
            return xml_input
        else:
            return None

    # def ensure_experiments(self, exp_no=None):
    #     if self.get_bbo_steps() is None:
    #         exp_count = len(self._var_coll.var_confs)
    #         if exp_no is not None:
    #             exp_count = exp_no
    #         if exp_count == 0:
    #             exp_count = 1
    #     else:
    #         exp_count = self.get_bbo_steps()
    #     self.get_files().ensure_exp_paths(exp_count)
    #
    # def load_libraries(self):
    #     lib_paths = []
    #     lib_elems = self._xml_input.xpath('/librec-auto/library')
    #     for elem in lib_elems:
    #         self._libraries.add_lib(
    #             Library(elem.text, elem.get('src'), self._files))

    # # Process config takes the config file and produces a dictionary of the following form:
    # # xpath-string => list of values
    # # or xpath-string => (range-to, range-from) pair
    # # Right now, we will assume the first
    # def process_config(self):
    #     self.setup_bbo()
    #     self._var_data = defaultdict(list)
    #     self.substitute_library()
    #     self.collect_vars()
    #     self.ensure_experiments()

    # def setup_bbo(self):
    #     opt_elem = single_xpath(self._xml_input, '/librec-auto/optimize')
    #     if opt_elem is None:
    #         self._bbo_steps = 0
    #     else:
    #         self._bbo_steps = int(single_xpath(opt_elem, 'iterations').text)

    # # Have to wait to writes experiment-specific XML configurations to each exp directory
    # # in case a purge is happening.
    # def setup_exp_configs(self, startflag=None):
    #     self.write_exp_configs(startflag)
    #
    # def substitute_library(self):
    #     ref_elems = self._xml_input.xpath('//*[@ref]')
    #     pro_feat_names = self.protected_features.get_protected_feature_names()
    #     for ref_elem in ref_elems:
    #         ref_name = ref_elem.get('ref')
    #         named_elem = self._libraries.get_elem(ref_name)
    #         if named_elem is not None:
    #             merged_elem = merge_elements(named_elem, ref_elem)
    #             ref_elem.getparent().replace(ref_elem, merged_elem)
    #         else:
    #             if ref_name in pro_feat_names:
    #                 continue
    #             else:
    #                 logging.warning(f"No such element in library {ref_name}")
    #
    # def collect_vars(self):
    #     self.collect_librec_vars()
    #     self.collect_rerank_vars()
    #     self._var_coll.compute_var_configurations()
    #
    # def collect_librec_vars(self):
    #     Tag = 'value'
    #
    #     value_elems = self._xml_input.xpath(
    #         '/librec-auto/*[not(self::rerank)]/*/value')
    #
    #     value_optimize_elems = self._xml_input.xpath(
    #         '/librec-auto/alg/*//lower')
    #
    #     check_multiple_values = [elem.getparent() for elem in value_elems]
    #     check_multiple_values = list(set(check_multiple_values))
    #
    #     has_optimize = len(self._xml_input.xpath('/librec-auto/optimize')) > 0
    #
    #     if len(check_multiple_values) != len(value_elems) and len(value_optimize_elems) > 0:
    #         raise InvalidConfiguration("optimization", "You may only use upper/lower for optimizing ranges")
    #     elif len(value_optimize_elems) > 0 and not has_optimize:
    #         raise InvalidConfiguration("optimization", "You may only use upper/lower for optimizing ranges")
    #     elif len(value_elems) > 0 and has_optimize:
    #         raise InvalidConfiguration("optimization", "Use of upper/lower value bounds requires optimize tag")
    #     elif len(value_optimize_elems) > 0:
    #         value_elems = value_optimize_elems + self._xml_input.xpath(
    #             '/librec-auto/alg/*//upper')
    #         parents = [elem.getparent() for elem in value_elems]
    #
    #         Tag = 'lower'
    #
    #     parents = [elem.getparent() for elem in value_elems]
    #     parents = list(set(parents))
    #     if Tag == 'value':
    #         for parent in parents:
    #             vals = [elem.text for elem in parent.iterchildren(tag=Tag)]
    #             parent_path = build_parent_path(parent)
    #             self._var_coll.add_var('librec', parent_path, vals)
    #
    #     else:
    #         for parent in parents:
    #             val_lower = [elem.text for elem in parent.iterchildren(tag='lower')]
    #             val_upper = [elem.text for elem in parent.iterchildren(tag='upper')]
    #             vals = []
    #             # print(val_lower,val_upper)
    #             for i in range(len(val_lower)):
    #                 vals.append(val_lower[i])
    #                 vals.append(val_upper[i])
    #
    #             parent_path = build_parent_path(parent)
    #             self._var_coll.add_var('librec', parent_path, vals)
    #
    # def collect_rerank_vars(self):
    #     value_elems = self._xml_input.xpath('/librec-auto/rerank/*//value')
    #     parents = [elem.getparent() for elem in value_elems]
    #     parents = list(set(parents))
    #     for parent in parents:
    #         vals = [elem.text for elem in parent.iterchildren(tag='value')]
    #         parent_path = build_parent_path(parent)
    #         self._var_coll.add_var('rerank', parent_path, vals)
    #
    # # Write versions of the config file in which the parameters with multiple values are replaced with
    # # a single value
    # def write_exp_configs(self, startflag=None, val=None, iteration=None):
    #
    #     configs = list(
    #         zip(self.get_files().get_exp_paths_iterator(),
    #             iter(self._var_coll.var_confs)))
    #     if self.get_bbo_steps() is not None and startflag is None:
    #         exp, vconf = configs[0]
    #
    #         for x in range(len(val)):
    #             vconf.vars[x].val = val[x]
    #         vconf.exp_no = None
    #         vconf.exp_dir = exp.exp_name
    #
    #         current_exp_path = self._files.get_study_path() / self._files.get_exp_name(iteration)
    #
    #         exp.set_path('conf', current_exp_path / 'conf')
    #         exp.exp_name = self._files.get_exp_name(iteration)
    #         self.write_exp_config(exp, vconf, current_exp_path)
    #
    #
    #     elif startflag is not None:
    #         i = 0
    #         for exp, vconf in configs[:1]:
    #             vconf.exp_no = i
    #             vconf.exp_dir = exp.exp_name
    #             self.write_exp_config(exp, vconf)
    #     else:
    #         i = 0
    #         for exp, vconf in configs:
    #             vconf.exp_no = i
    #             vconf.exp_dir = exp.exp_name
    #             self.write_exp_config(exp, vconf)
    #
    # def write_exp_config(self, exp, vconf, iteration=None):
    #     new_xml = copy.deepcopy(self._xml_input)
    #     # Remove libraries. All substitutions have already happened.
    #     for lib in new_xml.xpath('/librec-auto/library'):
    #         lib.getparent().remove(lib)
    #
    #     pat = re.compile(ConfigCmd._PARAM_NAME_PATH_RE)
    #     for vinfo in vconf.vars:
    #         var_elem = new_xml.xpath(vinfo.path)[0]
    #         attr_save = copy.copy(var_elem.attrib)
    #         var_elem.clear()
    #         var_elem.text = str(vinfo.val)
    #         var_elem.set("var", "true")
    #         if var_elem.tag == 'param':  # params are distinguished by name attributes
    #             mat = pat.match(vinfo.path)
    #             var_elem.attrib['name'] = mat.group(1)
    #         for key in attr_save.keys():
    #             var_elem.attrib[key] = attr_save[key]
    #     new_xml = self.protected_features.replace_referenced_protected_features(new_xml)
    #     new_xml.append(
    #         etree.Comment(
    #             'This configuration file was automatically generated by librec-auto. '
    #             +
    #             'Editing may produce unpredictable results and is not recommended.'
    #         ))
    #     outpath = exp.get_path('conf') / Files.DEFAULT_CONFIG_FILENAME
    #     logging.info('Writing config file ' + str(outpath))
    #     new_xml.getroottree().write(outpath.absolute().as_posix(),
    #                                 pretty_print=True)
    #
    #     props = LibrecProperties(new_xml, self._files)
    #     exp.add_to_config(props.properties, 'librec_result')
    #
    #     if iteration is not None:
    #         props.properties['dfs.result.dir'] = exp.exp_name + '/result'
    #
    #     props.save(exp)
    #
    #     if vconf.ref_config:
    #         path = exp.get_ref_exp_flag_path()
    #         with path.open(mode='w') as fh:
    #             fh.write(vconf.ref_config.exp_dir)
    #             fh.write('\n')
    #
    # def has_rerank(self):
    #     rerank_elems = self._xml_input.xpath('/librec-auto/rerank')
    #     return len(rerank_elems) > 0
    #
    # def has_post(self):
    #     post_elems = self._xml_input.xpath('/librec-auto/post')
    #     return len(post_elems) > 0
    #
    # def cross_validation(self):
    #     model_elem = single_xpath(self._xml_input,
    #                               '/librec-auto/splitter/model')
    #     if model_elem.text == 'kcv':
    #         return int(model_elem.get('count'))
    #     else:
    #         return 1

    def is_valid(self):
        return self._xml_input is not None

    # def thread_count(self):
    #     thread_elems = self._xml_input.xpath('/librec-auto/thread-count')
    #     if len(thread_elems) == 0:
    #         return 1
    #     else:
    #         return int(thread_elems[0].text)

    # def get_python_metrics(self):
    #     """
    #     Gets the XML elements for python-side metrics
    #     """
    #     return self._xml_input.findall('metric/script[@lang="python3"]')
    #
    # def get_cv_directories(self, absolute=False) -> List[Path]:
    #     """
    #     Gets the list of cv directories as Path objects.
    #     """
    #     data_path = self._files.get_data_path()
    #
    #     # Should really dispatch through the Files object
    #     split_dir = data_path / Path('split')  # cv splits live here
    #
    #     dir_strings = os.listdir(split_dir)  # ['cv_1', 'cv_2', ...]
    #
    #     dir_pat = re.compile(self._CV_DIR_RE)
    #
    #     cv_dir_strings = list(filter(dir_pat.match, dir_strings))
    #
    #     dirs = [split_dir / cv_dir_string for cv_dir_string in cv_dir_strings]
    #
    #     if absolute:
    #         return [dir.absolute() for dir in dirs]
    #     else:
    #         return dirs

def read_config_file(config_file, target, log_filename=None):
    config = ConfigCmd(config_file, target, log_filename)
    # print(target)
    # if config.is_valid():
    #     print()
        # config.load_libraries()
        # config.process_config()
    # else:
    #     raise InvalidConfiguration("Configuration file load error", "There was an error loading the configuration file.")
    return config


import numpy as np
from .abstract_ranker import AbstractRanker
import sys

# Objective function: min 1/2 (X - U . V_T) + 1/2 * \lambda (U)^2 + 1/2 * \lambda (V)^2

class FactorUCB(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(FactorUCB, self).__init__(config, dataObj)
        self.n_samples = np.zeros(dataObj.n_users)
        self.n_clicks = np.zeros(dataObj.n_users)
        self.prng = np.random.RandomState(seed=config.seed)
        self.sigma = float(parameters["sigma"]["value"]) if "sigma" in parameters else 1.0

        self.l = int(parameters["latent_dim"]["value"]) if "latent_dim" in parameters else 0
        self.lambda_1 = float(parameters["lambda1"]["value"]) if "lambda1" in parameters else 1.0
        self.lambda_2 = float(parameters["labmda2"]["value"]) if "lambda2" in parameters else 1.0
        self.alpha_a = float(parameters["alpha_a"]["value"]) if "alpha_a" in parameters else 1.0
        self.alpha_u = float(parameters["alpha_u"]["value"]) if "alpha_u" in parameters else 1.0

        self.V = np.zeros((self.dataObj.n_items, self.l))
        self.k = 0
        self.contextual_var = bool(parameters["contextual_variable"]["value"]) if "contextual_variable" in parameters else False
        if self.contextual_var:
            self.X = self.dataObj.feature_data['train_item_topical_features']
            self.k = self.X.shape[1]
            self.XV = np.concatenate((self.X, self.V), axis=1)
            # self.XV_optimal = np.concatenate((self.X, self.dataObj.feature_data['test_item_latent_features']), axis=1)
        else:
            self.XV = self.V
            # self.XV_optimal = self.dataObj.feature_data['test_item_latent_features']
        self.Theta = np.zeros((self.dataObj.n_users, self.k + self.l))
        self.Theta_x = self.Theta[:, :self.k]
        self.Theta_v = self.Theta[:, self.k:]
        self.W = np.identity(n=self.dataObj.n_users)

        self.A = np.zeros((self.dataObj.n_users, (self.k+self.l)*self.dataObj.n_users, (self.k+self.l)*self.dataObj.n_users))
        # self.A = np.zeros((self.dataObj.n_users, self.lambda_1 * np.identity(n=(self.k+self.l)*self.dataObj.n_users)))
        self.AInv = np.zeros((self.dataObj.n_users, (self.k+self.l)*self.dataObj.n_users, (self.k+self.l)*self.dataObj.n_users))
        # self.AInv = np.zeros((self.dataObj.n_users, self.lambda_1 * np.identity(n=(self.k + self.l) * self.dataObj.n_users)))
        identity = self.lambda_1 * np.identity(n=(self.k+self.l)*self.dataObj.n_users)
        for i in range(self.A.shape[0]):
            self.A[i] = identity
            self.AInv[i] = np.linalg.inv(self.A[i])  # np.zeros((self.k+self.l,self.k+self.l))
        self.b = np.zeros((self.dataObj.n_users, (self.k+self.l)*self.dataObj.n_users))

        self.C = np.zeros((self.dataObj.n_items, self.l, self.l))
        self.CInv = np.zeros((self.dataObj.n_items, self.l, self.l))
        for i in range(self.C.shape[0]):
            self.C[i] = self.lambda_2 * np.identity(n=self.l)
            self.CInv[i] = np.linalg.inv(self.C[i])  # np.zeros((self.k+self.l,self.k+self.l))
        self.d = np.zeros((self.dataObj.n_items, self.l))

        self.ill_matrix_counter = 0
        # for ill inverse
        self.AInv_tmp = np.zeros((self.dataObj.n_users, (self.k+self.l)*self.dataObj.n_users, (self.k+self.l)*self.dataObj.n_users))
        self.b_tmp = np.zeros((self.dataObj.n_users, (self.k+self.l)*self.dataObj.n_users))
        self.CInv_tmp = np.zeros((self.dataObj.n_items, self.l, self.l))
        self.d_tmp = np.zeros((self.dataObj.n_items, self.l))

    def get_ranking(self, batch_users, sampled_items=None, round=None):
        """
        :param x: features
        :param k: number of positions
        :return: ranking: the ranked item id.
        """
        # assert x.shape[0] >= k
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        # self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.dim))
        tie_breaker = self.prng.rand(len(sampled_items))
        for i in range(len(batch_users)):
            user = batch_users[i]

            # compute vec((X_a_t,V_a_t)W^T) -> N * (k+l)N
            XVW = self.create_vectorized_matrix(self.W.shape[0], self.W.shape[0]*(self.k+self.l), self.XV, self.W, range(len(sampled_items)))

            # compute line 9 of Algorithm 1
            score = np.dot(self.XV[sampled_items], np.dot(self.Theta.T, self.W[user].T))
            cb1 = np.sqrt(np.sum(np.multiply(np.dot(XVW,self.AInv[user]),XVW),axis=1))
            Theta_v_W = np.dot(self.Theta_v.T,self.W[user].T)
            var2 = [(np.dot(np.dot(Theta_v_W.T, self.CInv[item]),Theta_v_W)) for item in sampled_items]
            cb2 = np.sqrt(var2)
            ucb = score + self.alpha_u * cb1 + self.alpha_a * cb2

            selected_list = np.lexsort((tie_breaker, -ucb))[:self.config.list_size]
            rankings[i] = sampled_items[selected_list]

        return rankings

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        for i in range(len(batch_users)):
            user = batch_users[i]

            # compute vec((X_a_t,V_a_t)W^T) -> list_size * (k+l)N
            XVW = self.create_vectorized_matrix(self.W.shape[0], self.W.shape[0] * (self.k + self.l), self.XV, self.W, range(len(rankings[i])))
            # XVW_optimal = self.create_vectorized_matrix(self.W.shape[0], self.W.shape[0] * (self.k + self.l), self.XV_optimal[sampled_items], self.W, rankings[i])

            _clicks, _ranking, _XVW = self.__collect_feedback(clicks[i], rankings[i], XVW)

            # discount_coef = [1 / (math.log(1 + j)) for j in range(1, len(rankings[0]) + 1)]
            # discount_coef_reward = [math.log(1 + j) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [self.gamma * 1 / (math.log(1 + j)) for j in range(1, len(_clicks) + 1)]

            # if self.processing_type == 'recommended_discountfactor':
            #     self.exp_recommended[user][np.array(rankings[0])] += discount_coef
            # elif self.processing_type == 'examined_discountfactor':
            #     if len(clicks) == 0:
            #         self.exp_examined[user][np.array(rankings[0])] += discount_coef
            #     else:
            #         self.exp_examined[user][np.array(rankings[0][:len(clicks)])] += discount_coef[:len(clicks)]
            #
            # if self.processing_type == 'item_weight':
            #     _batch_features = self.update_item_weight(rankings[0], _batch_features, _clicks, discount_coef_penalization, discount_coef_reward, user, user_round)

            """
            This is for computing self.theta (Line 3-5 of Alogirthm 1 of NIPS 11)
            For fast matrix inverse, we use Woodbury matrix identity (https://en.wikipedia.org/wiki/Woodbury_matrix_identity)

            Return: self.theta is updated.
            """
            # for the inverse of M, feature matrix
            # XW * A^-1 * XW^T
            xAx = np.dot(_XVW, np.dot(self.AInv[user], _XVW.T))
            # (1/sigma I + xAx)^-1
            try:
                tmp_inv = np.linalg.inv(1 / self.sigma * np.eye(len(_XVW)) + xAx)
            except np.linalg.LinAlgError:
                # for the ill matrix. if the matrix is not invertible, we ignore this update
                self.ill_matrix_counter += 1
                return
            # A^-1*x^T
            AInv_xT = self.AInv[user].dot(_XVW.T)
            # AInv_xT*tmp_inv*AInv_xT^T
            self.AInv_tmp = np.dot(np.dot(AInv_xT, tmp_inv), AInv_xT.T)
            # MInv - the new part
            self.AInv[user] -= self.AInv_tmp
            self.A[user] += self.sigma * _XVW.T.dot(_XVW)

            # for b: feedback
            # if self.processing_type == 'feature_weight':
            #     self.update_feature_weight(_batch_features, _clicks, discount_coef_penalization, discount_coef_reward,
            #                                user, user_round)
            # else:
            self.b[user] += np.dot(_clicks, _XVW)
            # for parameter Theta
            self.Theta = self.devectorize(np.dot(self.AInv[user], self.b[user]), self.k+self.l)
            # self.theta[self.theta < 0] = 0
            self.Theta_x = self.Theta[:, :self.k]
            self.Theta_v = self.Theta[:, self.k:]


            # ranking = rankings[i][:len(_clicks)]
            Theta_v_W = np.dot(self.Theta_v.T, self.W[user].T)
            xx = np.dot(Theta_v_W.reshape(self.Theta_v.shape[1],1), Theta_v_W.reshape(self.Theta_v.shape[1],1).T)
            for i in range(len(_ranking)):
                item = _ranking[i]
                self.C[item] += xx
                self.CInv[item] = np.linalg.inv(self.C[item])
                if self.contextual_var:
                    # print('a='+str(self.d.shape)+', b='+str(Theta_v_W.shape)+', c='+str(self.X[ranking].shape)+', d='+str(self.Theta_x.T.shape)+', e='+str(self.W[user])+', f='+str((_clicks[i] - np.dot(self.X[ranking],np.dot(self.Theta_x.T, self.W[user]))).shape))
                    # sys.stdout.flush()
                    # clicked_items_index = _clicks[i].nonzero()[0]
                    self.d[item] += Theta_v_W * (_clicks[i] - np.dot(self.X[item],np.dot(self.Theta_x.T, self.W[user])))
                else:
                    self.d[item] += Theta_v_W * _clicks[i]
                self.V[item] = np.dot(self.CInv[item], self.d[item])
            self.XV[:, self.k:] = self.V

            self.n_samples[user] += len(_clicks)
            self.n_clicks[user] += sum(_clicks)

    def __collect_feedback(self, click, ranking, batch_X):
        """
        :param y:
        :return: the last observed position.
        """
        # With  Cascade assumption, only the first click counts.
        if self.config.feedback_model == 'cascade':
            if np.sum(click) == 0:
                return click, ranking, batch_X
            first_click = np.where(click)[0][0]
            # batch_X[ranking[:first_click+1]] = np.zeros((ranking[:first_click+1][0], batch_X.shape[1]))
            return click[:first_click + 1], ranking[:first_click + 1], batch_X[:first_click + 1]
        elif self.config.feedback_model == 'dcm':
            if np.sum(click) == 0:
                return click, batch_X
            last_click = np.where(click)[0][-1]
            return click[:last_click + 1], ranking[:last_click + 1], batch_X[:last_click + 1]
        # all items are observed
        else:
            return click, ranking, batch_X

    def create_vectorized_matrix(self, n_rows, n_cols, first_matrix, second_matrix, selected_rows):
        mat = np.zeros((n_rows, n_cols))
        mat = [self.vectorize(np.dot(self.matrixize(second_matrix.shape[0], first_matrix.shape[1], first_matrix[row], row).T,second_matrix)) for row in selected_rows]
        return np.array(mat)

    def matrixize(self, n_rows, n_cols, vec, target_row):
        mat = np.zeros((n_rows, n_cols))
        mat[target_row] = vec
        return mat

    def vectorize(self, M):
        # temp = []
        # for i in range(M.shape[0]*M.shape[1]):
        # 	temp.append(M.T.item(i))
        # V = np.asarray(temp)
        # return V
        return np.reshape(M.T, M.shape[0] * M.shape[1])

    def devectorize(self, Vec, dimension):
        # temp = np.zeros(shape = (C_dimension, len(V)/C_dimension))
        # for i in range(len(V)/C_dimension):
        # 	temp.T[i] = V[i*C_dimension : (i+1)*C_dimension]
        # W = temp
        # return W
        # To-do: use numpy built-in function reshape.
        return np.reshape(Vec, (int(len(Vec) / dimension), dimension))


from fair_dynamic_rec.core.rankers.random import *
from fair_dynamic_rec.core.rankers.naive_greedy import *
from fair_dynamic_rec.core.rankers.naive_epsilon_greedy import *
from fair_dynamic_rec.core.rankers.thompson_sampling import *
from fair_dynamic_rec.core.rankers.kl_upper_confidence_bound import *
from fair_dynamic_rec.core.rankers.linear_submodular_bandit import *
from fair_dynamic_rec.core.rankers.linear_upper_confidence_bound import *
from fair_dynamic_rec.core.rankers.hybrid_lsb_linucb import *
from fair_dynamic_rec.core.rankers.lsb_random import *
from fair_dynamic_rec.core.rankers.ea_linear_submodular_bandit import *
from fair_dynamic_rec.core.rankers.ea_linear_upper_confidence_bound import *
from fair_dynamic_rec.core.rankers.ea_hybrid_lsb_linucb import *
from fair_dynamic_rec.core.rankers.ears_linear_thompson_sampling import EARSLinearTS
from fair_dynamic_rec.core.rankers.ears_linear_submodular_bandit import EARSLSB
from fair_dynamic_rec.core.rankers.ears_linear_upper_confidence_bound import EARSLinUCB
from fair_dynamic_rec.core.rankers.ears_hybrid_lsb_linucb import EARSHybridLSBLinUCB
from fair_dynamic_rec.core.rankers.factor_ucb import FactorUCB
from fair_dynamic_rec.core.rankers.linear_neighbor_bandit import NeighborUCB
from tqdm import tqdm

RankerModel = {'random': Random, 'naive_greedy': NaiveGreedy, 'naive_epsilon_greedy': NaiveEpsilonGreedy,
               'thompson_sampling': ThompsonSampling, 'kl_upper_confidence_bound': KLUCB, 'linear_submodular_bandit': LSB,
               'linear_upper_confidence_bound': LinUCB, 'hybrid_lsb_linucb': HybridLSBLinUCB, 'lsb_random': LSBRandom,
               'ea_linear_submodular_bandit': EALSB, 'ea_linear_upper_confidence_bound': EALinUCB, 'ea_hybrid_lsb_linucb': EAHybridLSBLinUCB,
               'ears_linear_thompson_sampling': EARSLinearTS, 'ears_linear_submodular_bandit': EARSLSB,
               'ears_linear_upper_confidence_bound': EARSLinUCB, 'ears_hybrid_lsb_linucb': EARSHybridLSBLinUCB,
               'factor_ucb': FactorUCB, 'linear_neighbor_bandit': NeighborUCB}


def set_rankers(config, dataObj):
    rankers = []
    # print('config.rankers=' + str(len(config.rankers)))
    for params in tqdm(config.rankers):
        # print(params["single_val_params"]["name"]["value"])
        if len(params["multiple_val_params"]) > 0:
            for param in params["multiple_val_params"]:
                # print('params["multiple_val_params"]=' + str(len(params["multiple_val_params"])))
                _conf = dict(list(params["single_val_params"].items()) + list(param.items()))
                rankers.append(
                    {'ranker': RankerModel[params["single_val_params"]["name"]["value"]](config, dataObj, _conf),
                     'config': _conf}
                )
        elif len(params["single_val_params"]) > 0:
            rankers.append(
                {'ranker': RankerModel[params["single_val_params"]["name"]["value"]](config, dataObj, params["single_val_params"]),
                 'config': params["single_val_params"]}
            )
        # _ranker = RankerModel[params["single_val_params"]["name"]](config, dataObj)
        # # _ranker.load_data(config, dataObj)
        # _config = dict(list(config.items()) + list(params["single_val_params"].items()))
        # for param in params["multiple_val_params"]:
        #     _ranker.config = dict(list(_ranker.config.items()) + list(param.items()))
        #
        # rankers.append({'ranker': _ranker, 'config': _config})
    return rankers


import numpy as np
from .ea_linear_submodular_bandit import EALSB

class EALinUCBItemWeight(EALSB):
    def __init__(self, config, dataObj, parameters=None):
        super(EALinUCBItemWeight, self).__init__(config, dataObj, parameters)
        self.dim = self.dataObj.feature_data['train_item_latent_features'].shape[1]

        # parameters
        self.ill_matrix_counter = 0
        self.theta = np.ones((self.dataObj.n_users, self.dim))  # d-dimensional
        self.b = np.zeros((self.dataObj.n_users, self.dim))  # d
        self.M = np.zeros((self.dataObj.n_users, self.dim, self.dim))  # d by d
        self.MInv = np.zeros((self.dataObj.n_users, self.dim, self.dim))  # for fast matrix inverse computation, d by d
        for i in range(self.dataObj.n_users):
            self.M[i] = np.eye(self.dim)
            self.MInv[i] = np.eye(self.dim)
        # for ill inverse
        self.b_tmp = np.zeros((self.dataObj.n_users, self.dim))
        self.MInv_tmp = np.zeros((self.dataObj.n_users, self.dim, self.dim))

        self.X = np.zeros((self.dataObj.n_users, self.dataObj.feature_data['train_item_latent_features'].shape[0], self.dataObj.feature_data['train_item_latent_features'].shape[1]))
        for i in range(self.dataObj.n_users):
            self.X[i] = self.dataObj.feature_data['train_item_latent_features']
        self.gamma = float(parameters["gamma"]["value"])

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        """
        :param x: features
        :param k: number of positions
        :return: ranking: the ranked item id.
        """
        # assert x.shape[0] >= k
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.dim))
        tie_breaker = self.prng.rand(len(self.dataObj.feature_data['train_item_latent_features']))
        for i in range(len(batch_users)):
            user = batch_users[i]
            cb = self.alpha * np.sqrt(np.multiply(np.dot(self.X[user], self.MInv[user]), self.X[user]).sum(axis=1))
            score = np.dot(self.X[user], self.theta[user])
            ucb = score + cb
            rankings[i] = np.lexsort((tie_breaker, -ucb))[:self.config.list_size]
            self.batch_features[i] = self.X[user][rankings[i]]
        return rankings


import numpy as np
import os
from pathlib import Path
import json
from fair_dynamic_rec.core.util.utils import get_dict_key_from_value, get_dict_keys_from_list, get_param_config_name

def save_rankings_online_simulator(config, dataObj, params_name, round, rankings, clicks, users_batch):
    pre_path = make_output_dir(config, params_name)
    rankings_dict = {}
    for i in range(len(users_batch)):
        rankings_mapped_to_original = np.asarray(get_dict_keys_from_list(dataObj.itemid_mapped_data, rankings[i]))
        rankings_dict[str(get_dict_key_from_value(dataObj.userid_mapped_data, users_batch[i]))] = {'r': list(rankings_mapped_to_original), 'c': list(rankings_mapped_to_original[clicks[i].astype(np.bool)])}
    output_path = pre_path / Path(str(round))
    with open(output_path, 'w+') as f:
        json.dump(rankings_dict, f)

def save_rewards(config, rankers, overall_optimal_reward, overall_rewards):
    i = 0
    for ranker in rankers:
        pre_path = make_output_dir(config, get_param_config_name(ranker["config"]))
        output_path = pre_path / 'optimal-reward'
        with open(output_path, 'w+') as f:
            json.dump({'reward': list(overall_optimal_reward)}, f)
        output_path = pre_path / 'ranker-reward'
        with open(output_path, 'w+') as f:
            json.dump({'reward': list(overall_rewards[i])}, f)
        i += 1

def make_output_dir(config, params_name):
    if not os.path.exists(config._target / Path('results')):
        os.makedirs(config._target / Path('results'))
    if not os.path.exists(config._target / Path('results/' + params_name)):
        os.makedirs(config._target / Path('results/' + params_name))
    return config._target / Path('results/' + params_name)



from lxml import etree


# def build_parent_path(elem, pathsofar=''):
#     if elem.tag == 'param':  # params are distinguished by name attributes
#         pathsofar = f"[@name='{elem.get('name')}']"
#     nextpath = '/' + elem.tag + pathsofar
#     if (elem.getparent() == None):
#         return nextpath
#     elif (type(elem) is not etree._Element):
#         return nextpath
#     else:
#         return build_parent_path(elem.getparent(), nextpath)
#
#
# def read_xml_from_path_string(path_str):
#     path = Path(path_str)
#     if path.exists():
#         return xml_load_from_path(path)
#     else:
#         logging.error(f'XML file {path_str} does not exist. Load failed.')
#         return None


def xml_load_from_path(path):
    """
    Loads the XML file in a dictionary

    Prints a warning and returns an empty dictionary if the file can't be read.
    :param path: The file name
    :return: A dictionary with the XML rules
     """
    try:
        with path.open() as fd:
            txt = fd.read()
    except IOError as e:
        print("Error reading ", path)
        print("IO error({0}): {1}".format(e.errno, e.strerror))
        # logging.error("Error reading %s. IO error: (%d) %s", path, e.errno, e.strerror)
        return None

    return xml_load_from_text(txt)


def xml_load_from_text(txt):
    try:
        parser = etree.XMLParser(remove_comments=True)
        xml_data = etree.fromstring(txt, parser=parser)
        # xml_data = ET.fromstring(txt)
    except etree.XMLSyntaxError as e:
        print("Error parsing XML")
        print("LXML error in line: {0}".format(e.lineno))
        xml_data = None

    return xml_data


# # The idea is that we have a complete element lib_elem and a partial element mod_elem.
# # The mod_elem contents override those in lib_elem when they overlap.
# # Only works at the top-level of children in the element.
# def merge_elements(lib_elem, mod_elem):
#     lib = copy.deepcopy(lib_elem
#     mod = copy.deepcopy(mod_elem)
#
#     lib_tags = [child.tag for child in lib.iterchildren(tag=etree.Element)]
#     mod_tags = [child.tag for child in mod.iterchildren(tag=etree.Element)]
#
#     mod_extra = list(set(mod_tags).difference(set(lib_tags)))
#
#     for lchild in lib.iterchildren(tag=etree.Element):
#         mchild_lst = [child for child in mod.iterchildren(lchild.tag)]
#         if len(mchild_lst) > 0:
#             lchild.clear()
#             lchild.getparent().replace(lchild, mchild_lst[0])
#
#     for extra_tag in mod_extra:
#         matching = [child for child in mod.iterchildren(tag=extra_tag)]
#         lib.append(matching[0])
#
#     return lib
#
#
# def single_xpath(elem, path):
#     ans = elem.xpath(path)
#     if len(ans) > 0:
#         return ans[0]
#     else:
#         return None



import logging

class FairDynamicRecException(Exception):
    """Exceptions raised during the compilation of FairDynamicRec

    Attributes:
        element_name -- element which caused the error
        message -- explanation of the error
    """

    def __init__(self, elem_name, message="LibRec-Auto Exception."):
        self.element_name = elem_name
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.element_name} -> {self.message}'


class InvalidConfiguration(FairDynamicRecException):
    """Exception raised for errors in the configuration file XML.
    """

    # may want to make it less broad, go into more detail
    def __init__(self, elem_name, message="Error processing configuration file."):
        super().__init__(elem_name, message)
        self._write_to_log()

    def _write_to_log(self):
        logging.error(f' Configuration file: {self.element_name}: {self.message}')


import numpy as np
from .abstract_ranker import AbstractRanker

# Segment-based Epsilon-Greedy strategy: recommends playlists randomly with probability epsilon,
# otherwise recommends the top n_recos with highest mean observed rewards.
class NaiveEpsilonGreedy(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(NaiveEpsilonGreedy, self).__init__(config, dataObj)
        # self.user_segment = user_segment
        # n_segments = len(np.unique(self.user_segment))
        self.ranking_display = np.zeros((dataObj.n_users, dataObj.n_items))
        self.ranking_success = np.zeros((dataObj.n_users, dataObj.n_items))
        self.ranking_score = np.ones((dataObj.n_users, dataObj.n_items))
        self.epsilon = float(parameters["epsilon"]["value"])
        # self.cascade_model = cascade_model

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        # user_segment = np.take(self.user_segment, batch_users)
        user_scores = np.take(self.ranking_score, batch_users, axis = 0)
        user_random_score = np.random.random(user_scores.shape)
        n_users = len(batch_users)
        user_greedy = np.random.binomial(1, [1- self.epsilon for i in range(n_users)])
        new_scores = user_scores * user_greedy[:,np.newaxis]
        user_choice = np.lexsort((user_random_score, -new_scores))[:, :self.config.list_size]
        # Shuffle l_init first slots
        # np.random.shuffle(user_choice[0:l_init])
        return user_choice

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        batch_size = len(batch_users)
        for i in range(batch_size):
            # user_segment = self.user_segment[user_ids[i]]
            # total_stream = len(rewards[i].nonzero())
            nb_display = 0
            for r, c in zip(rankings[i], clicks[i]):
                nb_display +=1
                self.ranking_success[batch_users[i]][r] += c
                self.ranking_display[batch_users[i]][r] += 1
                self.ranking_score[batch_users[i]][r] = self.ranking_success[batch_users[i]][r] / self.ranking_display[batch_users[i]][r]
                # if self.cascade_model and ((total_stream == 0 and nb_display == l_init) or (r == 1)):
                #     break
        return


import numpy as np
from .abstract_ranker import AbstractRanker
import math

class EALSB(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(EALSB, self).__init__(config, dataObj)
        self.n_samples = np.zeros(dataObj.n_users)
        self.n_clicks = np.zeros(dataObj.n_users)
        self.dim = self.dataObj.feature_data['train_item_topical_features'].shape[1]
        self.prng = np.random.RandomState(seed=config.seed)
        self.alpha = float(parameters["alpha"]["value"])

        self.sigma = float(parameters["sigma"]["value"])
        self.ill_matrix_counter = 0
        self.theta = np.ones((self.dataObj.n_users, self.dim))  # d-dimensional
        self.b = np.zeros((self.dataObj.n_users, self.dim))  # d
        self.M = np.zeros((self.dataObj.n_users, self.dim, self.dim))  # d by d
        self.MInv = np.zeros((self.dataObj.n_users, self.dim, self.dim))  # for fast matrix inverse computation, d by d
        for i in range(self.dataObj.n_users):
            self.M[i] = np.eye(self.dim)
            self.MInv[i] = np.eye(self.dim)

        # for ill inverse
        self.b_tmp = np.zeros((self.dataObj.n_users, self.dim))
        self.MInv_tmp = np.zeros((self.dataObj.n_users, self.dim, self.dim))
        self.batch_features = None

        # self.X = np.zeros((self.dataObj.n_users, self.dataObj.feature_data['train_item_topical_features'].shape[0], self.dataObj.feature_data['train_item_topical_features'].shape[1]))
        # for i in range(self.dataObj.n_users):
        #     self.X[i] = self.dataObj.feature_data['train_item_topical_features']
        self.gamma = float(parameters["gamma"]["value"])
        self.type = parameters["type"]["value"]
        self.window = int(parameters['window']['value'])
        self.click_history = np.zeros((self.dataObj.n_users, self.dim))

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        """
        :param x: features
        :param k: number of positions
        :return: ranking: the ranked item id.
        delta: the conditional topic coverage of each item. Eq. (3) of NIPS 11 paper.
        """
        # assert x.shape[0] >= k
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.dim))
        tie_breaker = self.prng.rand(len(self.dataObj.feature_data['train_item_topical_features']))
        for i in range(len(batch_users)):
            user = batch_users[i]
            coverage = np.zeros(self.dim)
            ranking = []
            for j in range(self.config.list_size):
                # Line 8 - 11 of Nips 11
                gain_in_topic_coverage = self.conditional_coverage(x=self.dataObj.feature_data['train_item_topical_features'], coverage=coverage)
                cb = self.alpha * np.sqrt(np.multiply(np.dot(gain_in_topic_coverage, self.MInv[user]), gain_in_topic_coverage).sum(axis=1))
                score = np.dot(gain_in_topic_coverage, self.theta[user])
                ucb = score + cb + 1e-6 * tie_breaker

                winner = np.argmax(ucb)
                while winner in ranking:
                    ucb[winner] = -np.inf
                    winner = np.argmax(ucb)

                ranking.append(winner)
                self.batch_features[i][j] = gain_in_topic_coverage[winner]

                coverage = self.ranking_coverage(self.dataObj.feature_data['train_item_topical_features'][ranking])
            rankings[i] = np.asarray(ranking)
        return rankings

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        for i in range(len(batch_users)):
            user = batch_users[i]
            _clicks, _batch_features = self.__collect_feedback(clicks, i)

            discount_coef = [pow(self.gamma, j) for j in range(1,len(_clicks)+1)]
            #1
            # discount_coef_reward = [1 + pow(self.gamma, len(rankings[i])-j+1) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [x+1 for x in discount_coef]
            #2
            # discount_coef_reward = [pow(self.gamma, len(rankings[i]) - j + 1) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [x for x in discount_coef]
            #3
            # discount_coef_reward = [pow(self.gamma, len(rankings[i]) - j + 1) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [pow(1-self.gamma, j) for j in range(1, len(_clicks) + 1)]
            # # 4
            # discount_coef_reward = [1 + pow(self.gamma, len(rankings[i]) - j + 1) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [pow(1 - self.gamma, j) for j in range(1, len(_clicks) + 1)]
            # # 6
            # discount_coef_reward = [1 + math.log(j) for j in range(1, len(_clicks) + 1)]
            # 7
            discount_coef_reward = [1 + math.log(j) for j in range(1, len(_clicks) + 1)]
            discount_coef_penalization = [self.gamma * 1 / (1 + math.log(j)) for j in range(1, len(_clicks) + 1)]

            # if self.type == 'item_weight':
            #     clicked_items_index = np.where(_clicks == 1)[0]
            #     if len(clicked_items_index) == 0:
            #         self.X[user][rankings[i], :] = np.multiply(np.array(discount_coef).reshape(len(_clicks), 1), self.X[user][rankings[i], :])
            #     else:
            #         previous_clicked_item_index = 0
            #         for clicked_item_index in clicked_items_index:
            #             current_clicked_item_index = clicked_item_index
            #             self.X[user][rankings[i][current_clicked_item_index], :] = (2 - discount_coef[current_clicked_item_index]) * self.X[user][rankings[i][current_clicked_item_index], :]
            #             self.X[user][rankings[i][previous_clicked_item_index: current_clicked_item_index], :] = np.multiply(np.array(discount_coef[previous_clicked_item_index: current_clicked_item_index]).reshape(current_clicked_item_index-previous_clicked_item_index,1), self.X[user][rankings[i][previous_clicked_item_index: current_clicked_item_index], :])
            #             previous_clicked_item_index = current_clicked_item_index + 1

            """
            This is for computing self.theta (Line 3-5 of Alogirthm 1 of NIPS 11)
            For fast matrix inverse, we use Woodbury matrix identity (https://en.wikipedia.org/wiki/Woodbury_matrix_identity)

            Return: self.theta is updated.
            """
            # for the inverse of M, feature matrix
            # x * m^-1 * x^T
            xmx = np.dot(_batch_features, np.dot(self.MInv[user], _batch_features.T))
            # (1/sigma I + xmx)^-1
            try:
                tmp_inv = np.linalg.inv(1 / self.sigma * np.eye(len(_batch_features)) + xmx)
            except np.linalg.LinAlgError:
                # for the ill matrix. if the matrix is not invertible, we ignore this update
                self.ill_matrix_counter += 1
                return
            # m^-1*x^T
            MInv_xT = self.MInv[user].dot(_batch_features.T)
            # MInv_xT*tmp_inv*MInv_xT^T
            self.MInv_tmp[user] = np.dot(np.dot(MInv_xT, tmp_inv), MInv_xT.T)
            # MInv - the new part
            self.MInv[user] -= self.MInv_tmp[user]

            self.M[user] += self.sigma * _batch_features.T.dot(_batch_features)

            # for b: feedback
            # if self.type == 'feature_weight' and round % self.window == 0:
            #     clicked_items_index = np.where(_clicks == 1)[0]
            #     if len(clicked_items_index) == 0:
            #         _batch_features = -np.multiply(np.array(discount_coef_penalization).reshape(len(_clicks), 1), _batch_features)
            #     else:
            #         previous_clicked_item_index = 0
            #         for clicked_item_index in clicked_items_index:
            #             current_clicked_item_index = clicked_item_index
            #             _batch_features[current_clicked_item_index, :] = discount_coef_reward[current_clicked_item_index] * _batch_features[current_clicked_item_index, :]
            #             _batch_features[previous_clicked_item_index: current_clicked_item_index, :] = -np.multiply(np.array(discount_coef_penalization[previous_clicked_item_index: current_clicked_item_index]).reshape(current_clicked_item_index-previous_clicked_item_index,1), _batch_features[previous_clicked_item_index: current_clicked_item_index, :])
            #             previous_clicked_item_index = current_clicked_item_index + 1
            #     self.b_tmp[user] = _batch_features.sum(axis=0)
            # else:
            #     self.b_tmp[user] = np.dot(_clicks, _batch_features)
            if self.type == 'feature_weight':
                clicked_items_index = np.where(_clicks == 1)[0]
                _x = _batch_features
                if len(clicked_items_index) == 0:
                    _x = -np.multiply(np.array(discount_coef_penalization).reshape(len(_clicks), 1), _batch_features)
                else:
                    previous_clicked_item_index = 0
                    for clicked_item_index in clicked_items_index:
                        current_clicked_item_index = clicked_item_index
                        _x[current_clicked_item_index, :] = discount_coef_reward[current_clicked_item_index] * _batch_features[current_clicked_item_index, :]
                        _x[previous_clicked_item_index: current_clicked_item_index, :] = -np.multiply(np.array(discount_coef_penalization[previous_clicked_item_index: current_clicked_item_index]).reshape(current_clicked_item_index-previous_clicked_item_index,1), _batch_features[previous_clicked_item_index: current_clicked_item_index, :])
                        previous_clicked_item_index = current_clicked_item_index + 1
                self.click_history[user] += _x.sum(axis=0)
                self.b_tmp[user] = np.dot(_clicks, _batch_features)
                if user_round[user] % self.window == 0:
                    self.b_tmp[user] = self.click_history[user] / self.window
                    self.click_history[user] = np.zeros(self.dim)
            else:
                self.b_tmp[user] = np.dot(_clicks, _batch_features)
            self.b[user] += self.b_tmp[user]

            # for parameter theta
            self.theta[user] = np.dot(self.MInv[user], self.b[user])
            # self.theta[self.theta < 0] = 0

            self.n_samples[user] += len(_clicks)
            self.n_clicks[user] += sum(_clicks)

    def __collect_feedback(self, clicks, batch_user_id):
        """
        :param y:
        :return: the last observed position.
        """
        # With  Cascade assumption, only the first click counts.
        if self.config.feedback_model == 'cascade':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id], self.batch_features[batch_user_id]
            first_click = np.where(clicks[batch_user_id])[0][0]
            return clicks[batch_user_id][:first_click + 1], self.batch_features[batch_user_id][:first_click + 1]
        elif self.config.feedback_model == 'dcm':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id], self.batch_features[batch_user_id]
            last_click = np.where(clicks[batch_user_id])[0][-1]
            return clicks[batch_user_id][:last_click + 1], self.batch_features[batch_user_id][:last_click + 1]
        # all items are observed
        else:
            return clicks[batch_user_id], self.batch_features[batch_user_id]